import io
from datetime import datetime

import scrapy

from opennem.pipelines.nem import (
    ExtractCSV,
    ReadStringHandle,
    TableRecordSplitter,
    UnzipSingleFilePipeline,
)


def get_date_component(format_str):
    return datetime.now().strftime(format_str)


class NemXLSSpider(scrapy.Spider):

    start_url = ""

    url_params = {
        "day": get_date_component("%d"),
        "month": get_date_component("%m"),
        "year": get_date_component("%Y"),
    }

    def start_requests(self):
        request_url = self.start_url.format(**self.url_params)

        yield scrapy.Request(request_url)

    def parse(self, response):
        yield {"content": response.text}


class NemSingleMMSSpider(scrapy.Spider):

    pipelines = set(
        [
            UnzipSingleFilePipeline,
            ReadStringHandle,
            ExtractCSV,
            TableRecordSplitter,
        ]
    )

    def start_requests(self):
        if not hasattr(self, "url"):
            raise Exception("{} requires url parameter".format(self.__class__))

        yield scrapy.Request(self.url)

    def parse(self, response):
        yield {"body_stream": io.BytesIO(response.body)}



from typing import List

from opennem.core.normalizers import normalize_duid
from opennem.schema.time import TimeInterval, TimePeriod


def station_id_case(station_codes: List[str]) -> str:
    return ",".join(
        ["'{}'".format(i) for i in map(normalize_duid, station_codes)]
    )


def observation_query(
    station_codes: List[str],
    interval: TimeInterval,
    period: TimePeriod,
    timezone: str = "AEST",
):
    __query = """with intervals as (
        select generate_series(
            date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
            date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
            '{interval}'::interval
        )::timestamp as interval
    )

    select
        i.interval,
        fs.generated,
        fs.station_id
    from intervals i
    left outer join
        (select
            date_trunc('{trunc}', fs.observation_time AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
            fs.station_id,
            coalesce(avg(fs.temp_air), 0) as generated
            from bom_observation fs
            where
                fs.station_id in ({station_codes})
                and fs.observation_time > now() AT TIME ZONE '{timezone}' - '{period}'::interval
            group by 1, 2
        ) as fs on fs.interval = i.interval
    order by 1 desc, 2 desc"""

    query = __query.format(
        station_codes=station_id_case(station_codes),
        trunc=interval.trunc,
        interval=interval.interval_human,
        interval_remainder=interval.get_sql_join(
            field="observation_time", timezone=timezone
        ),
        period=period.period_sql,
        timezone=timezone,
    )

    return query



from urllib.parse import urlparse


def replace_database_in_url(db_url, db_name):
    """
        replaces the database portion of a database connection URL with db_name

        @param db_name database name to replace with
    """

    db_url_parsed = urlparse(db_url)
    db_url_parsed = db_url_parsed._replace(path=f"/{db_name}")

    return db_url_parsed.geturl()



from fastapi import HTTPException
from starlette import status

from opennem.core.time import (
    INTERVALS_SUPPORTED,
    PERIODS_SUPPORTED,
    get_interval,
    get_period,
)
from opennem.schema.time import TimeInterval


def human_to_interval(interval_human: str) -> TimeInterval:
    """
        Parses user supplied intervals like "15M" to TimeInterval
        instances.

    """
    if interval_human not in INTERVALS_SUPPORTED:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Interval not supported. Select one of: {}".format(
                ", ".join(INTERVALS_SUPPORTED)
            ),
        )

    return get_interval(interval_human)


def human_to_period(period_human: str) -> TimeInterval:
    if period_human not in PERIODS_SUPPORTED:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Period not supported. Select one of: {}".format(
                ", ".join(PERIODS_SUPPORTED)
            ),
        )

    return get_period(period_human)




from opennem.pipelines.wem.balancing_summary import WemStoreBalancingSummary
from opennem.spiders.wem import WemCurrentSpider, WemHistoricSpider


class WemCurrentBalancingSummary(WemCurrentSpider):
    name = "au.wem.current.balancing_summary"
    start_url = "http://data.wa.aemo.com.au/public/public-data/datafiles/balancing-summary/balancing-summary-{year}.csv"

    pipelines_extra = set([WemStoreBalancingSummary])


class WemHistoricBalancingSummary(WemHistoricSpider):
    name = "au.wem.historic.balancing_summary"
    start_url = "http://data.wa.aemo.com.au/public/public-data/datafiles/balancing-summary/"
    limit = 0

    pipelines_extra = set([WemStoreBalancingSummary])



from datetime import date, datetime
from decimal import Decimal
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, validator

from opennem.schema.core import BaseConfig
from opennem.schema.time import TimeIntervalAPI, TimePeriodAPI


class ScadaInterval(object):
    date: datetime
    value: Optional[float]

    def __init__(self, date: datetime, value: Optional[float] = None):
        self.date = date
        self.value = value


class ScadaReading(Tuple[datetime, Optional[float]]):
    pass


class UnitScadaReading(BaseModel):
    code: str
    data: List[ScadaReading]


class StationScadaReading(BaseModel):
    code: str
    facilities: Dict[str, UnitScadaReading]


class OpennemDataHistory(BaseConfig):
    start: Union[datetime, date]
    last: Union[datetime, date]
    interval: str
    data: List[Optional[float]]

    @validator("data")
    def validate_data(cls, data_value):
        data_value = list(
            map(lambda x: round(x, 2) if x else None, data_value)
        )

        return data_value


class OpennemData(BaseConfig):
    id: Optional[str]
    type: Optional[str]
    region: Optional[str]
    fuel_tech: Optional[str]

    network: Optional[str]
    data_type: str
    code: str = ""
    units: str

    interval: TimeIntervalAPI
    period: TimePeriodAPI

    history: OpennemDataHistory


class OpennemDataSet(BaseConfig):
    network: Optional[str]
    code: str
    region: Optional[str]
    created_at: Optional[datetime]

    data: List[OpennemData]


class DataQueryResult(BaseConfig):
    interval: datetime
    result: Union[float, int, None, Decimal]
    group_by: Optional[str]



from opennem.core.loader import load_data

SINGLE_UNITS = load_data("single_units.json")


def facility_unit_numbers_are_single(facility_duid: str) -> bool:
    """
        These units have unit numbers that are single units rather than ranges.

        ex. GT 1-2 means unit of alias GT1 and id 2 rather than alias GT and range 1-2

    """

    if not type(SINGLE_UNITS) is list:
        raise Exception("Error loading facility station map: not a list")

    return facility_duid in SINGLE_UNITS



import json
import logging
from datetime import datetime
from itertools import groupby

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.sql import text

from opennem.core.dispatch_type import DispatchType, parse_dispatch_type
from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_aemo_region,
    normalize_duid,
    participant_name_filter,
    station_name_cleaner,
)
from opennem.core.station_duid_map import (
    facility_has_station_remap,
    facility_map_station,
)
from opennem.core.unit_codes import get_unit_code
from opennem.core.unit_parser import parse_unit_duid
from opennem.db.models.opennem import (
    Facility,
    FacilityStatus,
    Participant,
    Station,
)
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def has_unique_duid(units: list) -> bool:
    duids = set([i["duid"] for i in units])
    return len(duids) == len(units)


def get_unique_duid(units: list) -> str:
    if not type(units) is list or len(units) < 1:
        return None

    first_record = units[0]

    return (
        normalize_duid(first_record["duid"])
        if "duid" in first_record
        else None
    )


def get_station_record_from_facilities(units: list):
    if not type(units) is list or len(units) < 1:
        raise Exception("Passed units list with no units ..")

    if len(units) == 1:
        return units[0]

    for u in units:
        cap = clean_capacity(u["reg_cap"])
        duid = normalize_duid(u["duid"])

        if cap:
            return u

    return units[0]


def get_unique_reqion(units: list) -> str:
    if not type(units) is list or len(units) < 1:
        return None

    first_record = units[0]

    return first_record["region"].strip()


class RegistrationExemptionGrouperPipeline(object):
    """
        This is the first-pass pipeline of the AEMO Registration and Exemptions list

        It groups by clean station name

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if not "generators" in item:
            raise Exception("No generators found in item pipeline failed")

        generators = item["generators"]

        # Add clean station names and if group_by name
        generators = [
            {
                **i,
                "name": station_name_cleaner(i["station_name"]),
                "name_join": False
                if facility_station_join_by_name(
                    station_name_cleaner(i["station_name"])
                )
                else i["duid"],
            }
            for i in generators
        ]

        # sort by name

        generators_grouped = {}

        for k, v in groupby(
            generators, key=lambda v: (v["name"], v["name_join"])
        ):
            key = k
            if not key in generators_grouped:
                generators_grouped[key] = []

            generators_grouped[key] += list(v)

        # with open("registrtation-exemption-grouped.json", "w") as fh:
        #     json.dump(generators_grouped, fh, indent=4)

        return {**item, "generators": generators_grouped}


class RegistrationExemptionStorePipeline(DatabaseStoreBase):
    """
        This pipeline is the second pass of the AEMO NEM REL List

        It sorts out whats new, what is updated, units and updating the database

    """

    # Cache of participant keys
    participant_keys = {}

    def process_participants(self, participants):
        s = self.session()

        for participant_data in participants:
            participant_name = participant_data["name"].strip()

            try:
                participant = (
                    s.query(Participant)
                    .filter(Participant.name == participant_name)
                    .one_or_none()
                )
            except MultipleResultsFound as e:
                logger.info(
                    "Found multiple participants with name {}".format(
                        participant_name
                    )
                )

                participant = (
                    s.query(Participant)
                    .filter(Participant.name == participant_name)
                    .first()
                )

            if not participant:
                participant = Participant(name=participant_name)

            participant.name_clean = participant_name_filter(participant_name)
            participant.abn = participant_data["abn"]

            s.add(participant)

            logger.info(
                "Found new NEM participant: {}".format(participant_name)
            )
            self.participant_keys[participant_name] = participant

            facility = None

        s.commit()

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        if not "generators" in item and type(item["generators"]) is not list:
            raise Exception("Invalid item - no generators located")

        if (
            not "participants" in item
            and type(item["participants"]) is not list
        ):
            raise Exception("Invalid item - no generators located")

        generators = item["generators"]
        participants = item["participants"]

        if not type(generators) is dict:
            raise Exception(
                "Require grouped generators for this pipeline step. Run RegistrationExemptionGrouperPipeline"
            )

        self.process_generators(generators)

    def process_generators(self, generators):
        s = self.session()

        stations_updated = 0
        stations_added = 0
        generators_updated = 0
        generators_added = 0

        for station_key, facilities in generators.items():
            facility = None
            facility_station = None
            created_station = False
            created_facility = False

            station_name = station_key[0]

            duid_unique = has_unique_duid(facilities)
            facility_count = len(facilities)

            # Step 1. Find the station
            # First by duid if it's unique
            duid = get_unique_duid(facilities)

            # This is the most suitable unit record to use for the station
            # see helper above
            facility_station_record = get_station_record_from_facilities(
                facilities
            )

            facility_network_region = get_unique_reqion(facilities)

            if duid and duid_unique and facility_count == 1:
                facility_lookup = None

                try:
                    facility_lookup = (
                        s.query(Facility)
                        .filter(Facility.network_code == duid)
                        .one_or_none()
                    )
                except MultipleResultsFound:
                    logger.warning(
                        "REL: Multiple stations found for {} {} with duid {}".format(
                            station_name, facility_network_region, duid
                        )
                    )

                if facility_lookup and facility_lookup.station:
                    facility_station = facility_lookup.station

            if (
                duid
                and (duid_unique and facility_count > 1)
                or not duid_unique
            ):

                facility_lookup = (
                    s.query(Facility)
                    .filter(Facility.network_code == duid)
                    .first()
                )

                if facility_lookup and facility_lookup.station:
                    facility_station = facility_lookup.station

            if not facility_station and facility_station_join_by_name(
                station_name
            ):
                try:
                    facility_station = (
                        s.query(Station)
                        .filter(Station.name == station_name)
                        .one_or_none()
                    )
                except MultipleResultsFound:
                    logger.warning(
                        "REL: Multiple stations found for {} {}".format(
                            station_name, facility_network_region
                        )
                    )

                logger.debug(
                    "REL: Looked up {} by name and region {} and found {}".format(
                        station_name,
                        facility_network_region,
                        facility.station if facility else "nothing",
                    )
                )

            # Create one as it doesm't exist
            if not facility_station:
                facility_station = Station(
                    name=station_name,
                    network_name=name_normalizer(
                        facility_station_record["station_name"]
                    ),
                    network_id="NEM",
                    created_by="pipeline.aemo.registration_exemption",
                )

                s.add(facility_station)
                created_station = True
            else:
                facility_station.updated_by = (
                    "pipeline.aemo.registration_exemption"
                )

            logger.info(
                "REL: {} station with name {} and code {}".format(
                    "Created" if created_station else "Updated",
                    facility_station.name,
                    facility_station.code,
                )
            )

            # Step 2. Add the facilities/units to the station
            # Now that we have a station or created one ..

            # Step go through the facility records we got ..
            for facility_record in facilities:
                network_name = name_normalizer(facility_record["station_name"])
                participant_name = name_normalizer(
                    facility_record["participant"]
                )
                facility_region = normalize_aemo_region(
                    facility_record["region"]
                )
                duid = normalize_duid(facility_record["duid"])
                reg_cap = clean_capacity(facility_record["reg_cap"])
                unit = parse_unit_duid(facility_record["unit_no"], duid)
                unit_size = clean_capacity(facility_record["unit_size"])
                unit_code = get_unit_code(
                    unit, duid, facility_station_record["station_name"]
                )
                facility_status = "operating"
                facility_dispatch_type = parse_dispatch_type(
                    facility_record["dispatch_type"]
                )
                fueltech = lookup_fueltech(
                    facility_record["fuel_source_primary"],
                    facility_record["fuel_source_descriptor"],
                    facility_record["tech_primary"],
                    facility_record["tech_primary_descriptor"],
                    facility_record["dispatch_type"],
                )

                # Skip loads that are not batteries or pumps for now
                # @NOTE @TODO better to centralize this as it needs to be consistent
                if (
                    facility_dispatch_type == DispatchType.LOAD
                    and fueltech not in ["battery_charging", "pumps",]
                ):
                    continue

                # check if we have it by ocode first
                facility = (
                    s.query(Facility)
                    .filter(Facility.code == unit_code)
                    .one_or_none()
                )

                # If the duid is unique then we have no issues on which to join/create
                if duid and duid_unique and not facility:
                    try:
                        facility = (
                            s.query(Facility)
                            .filter(Facility.network_code == duid)
                            .one_or_none()
                        )
                    except MultipleResultsFound:
                        logger.warning(
                            "REL: Multiple facilities found for {} {}".format(
                                station_name, duid
                            )
                        )

                if duid and not duid_unique and not facility:
                    facility = (
                        s.query(Facility)
                        .filter(Facility.network_code == duid)
                        .filter(Facility.unit_number == None)
                        .filter(Facility.status_id == "operating")
                        .first()
                    )

                # If the duid is not unique then we need to figure things out ..
                if duid and not duid_unique and not facility:
                    facility_lookup = (
                        s.query(Facility)
                        .filter(Facility.network_code == duid)
                        # Not having a code means we haven't written to this record yet so we'll use it
                        .filter(Facility.code == None)
                        .all()
                    )

                    facility_db_count = len(facility_lookup)

                    logging.debug(
                        "Non unique duid: {} with {} in database and {} in facility duid is {}".format(
                            station_name,
                            facility_db_count,
                            facility_count,
                            duid,
                        )
                    )

                    if len(facility_lookup) > 0:
                        facility = facility_lookup.pop()

                if not facility:
                    facility = Facility(
                        code=unit_code,
                        network_code=duid,
                        created_by="pipeline.aemo.registration_exemption",
                    )
                    created_facility = True
                else:
                    facility.updated_by = (
                        "pipeline.aemo.registration_exemption"
                    )

                # Sanity checking
                if len(unit_code) < 3:
                    raise Exception(
                        "Unit code {} is invalid. For station {} with duid {}".format(
                            unit_code, station_name, duid
                        )
                    )

                #
                if not facility.code:
                    facility.code = unit_code

                facility.fueltech_id = fueltech
                facility.network_code = duid
                facility.network_region = facility_region
                facility.network_name = network_name

                facility.capacity_registered = reg_cap
                facility.dispatch_type = facility_dispatch_type

                facility.unit_id = unit.id
                facility.unit_number = unit.number
                facility.unit_alias = unit.alias
                facility.unit_capacity = unit_size

                # Assume all REL's are operating if we don't have a status
                facility.status_id = "operating"

                facility.station = facility_station

                # Log that we have a new fueltech
                if fueltech and fueltech != facility.fueltech_id:
                    logger.warning(
                        "Fueltech mismatch for {} {}: prev {} new {}".format(
                            facility.name_clean,
                            facility.code,
                            facility.fueltech_id,
                            fueltech,
                        )
                    )

                if not created_facility:
                    facility.updated_by = (
                        "pipeline.aemo.registration_exemption"
                    )

                s.add(facility)
                s.commit()

                logger.info(
                    "REL: {} facility with duid {} and id {}".format(
                        "Created" if created_facility else "Updated",
                        facility.code,
                        facility.network_code,
                    )
                )

            generators_updated += 1

        logger.info(
            "NEM REL Pipeline: Added {} stations, updated {} stations. Added {}, updated {} generators of {} total".format(
                stations_added,
                stations_updated,
                generators_added,
                generators_updated,
                len(generators),
            )
        )



from typing import Optional

from pydantic import BaseSettings, PostgresDsn, RedisDsn
from scrapy.settings import Settings
from scrapy.utils.project import get_project_settings


class OpennemSettings(BaseSettings):
    env: str = "development"
    db_url: PostgresDsn = "postgresql://opennem:opennem@127.0.0.1:15433/opennem"
    cache_url: RedisDsn = "redis://127.0.0.1"
    sentry_url: Optional[str]

    google_places_api_key: str
    requests_cache_path: str = ".requests"

    slack_hook_url: Optional[str]

    s3_bucket_path: str = "s3://data.opennem.org.au/v3/"

    interval_default: str = "15m"
    period_default: str = "7d"

    # @todo overwrite scrapy settings here
    scrapy: Optional[Settings] = get_project_settings()

    class Config:
        env_file = ".env"
        fields = {
            "env": {"env": "ENV"},
            "db_url": {"env": "DATABASE_HOST_URL"},
            "cache_url": {"env": "REDIS_HOST_URL"},
            "sentry_url": {"env": "SENTRY_URL"},
            "slack_hook_url": {"env": "WATCHDOG_SLACK_HOOK"},
        }



from io import BytesIO

import scrapy
from openpyxl import load_workbook

from opennem.pipelines.aemo.general_information import (
    GeneralInformationGrouperPipeline,
    GeneralInformationStoragePipeline,
)


class AEMOGeneralInformationCurrentSpider(scrapy.Spider):
    """
        Extracts station and unit data from the AEMO general information spreadsheet


    """

    name = "au.aemo.current.general_information"

    start_urls = [
        # "https://aemo.com.au/-/media/files/electricity/nem/planning_and_forecasting/generation_information/nem-generation-information-april-2020.xlsx?la=en",
        # "https://data.opennem.org.au/v3/data/NEM+Generation+Information+April+2020.xlsx",
        "https://data.opennem.org.au/v3/data/NEM+Generation+Information+July+2020.xlsx"
    ]

    keys = [
        "Region",
        "Status",
        "station_name",
        "Owner",
        "TechType",
        "FuelType",
        "duid",
        "Units",
        "unit_capacity_lower",
        "unit_capacity",
        "UpperCapacity",
        "NameCapacity",
        "StorageCapacity",
        "UnitStatus",
        "DispatchType",
        "UseDate",
        "ClosureDateExpected",
        "ClosureDate",
        "SurveyID",
        "SurveyVersion",
        "SurveyEffective",
    ]

    pipelines_extra = set(
        [GeneralInformationGrouperPipeline, GeneralInformationStoragePipeline]
    )

    def parse(self, response):
        wb = load_workbook(BytesIO(response.body), data_only=True)

        ws = wb.get_sheet_by_name("ExistingGeneration&NewDevs")

        records = []

        for row in ws.iter_rows(min_row=3, values_only=True):

            # pick out the columns we want
            # lots of hidden columns in the sheet
            row_collapsed = row[0:10] + (row[11],) + row[12:19] + row[21:]

            return_dict = dict(zip(self.keys, list(row_collapsed)))

            if return_dict is None:
                raise Exception("Failed on row: {}".format(row))

            records.append(return_dict)

        yield {"records": records}



import logging
import re
from datetime import datetime
from pprint import pprint
from typing import List, Optional, Union

from dictalchemy.utils import fromdict
from pydantic import BaseModel
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm.session import make_transient

from opennem.core.loader import load_data
from opennem.db import SessionLocal, engine
from opennem.db.initdb import init_opennem
from opennem.db.load_fixtures import load_fixtures
from opennem.db.models.opennem import Facility, Location, Revision, Station
from opennem.importer.aemo_gi import gi_import
from opennem.importer.aemo_rel import rel_import
from opennem.importer.mms import mms_import
from opennem.importer.registry import registry_import
from opennem.schema.opennem import StationSchema

from .comparator import compare_record_differs

logger = logging.getLogger(__name__)
s = SessionLocal()


def load_revisions(stations):
    pass


def create_revision(model):
    # s = session()

    # make_transient(model)

    clone = model.asdict(exclude_pk=True)
    # clone.pop("id")
    # model.fromdict(clone)

    # pylint: disable=no-member
    clone_model = Station().fromdict(clone)

    return clone_model


def get_schema_name(record: object) -> str:
    class_name = str(record.__class__.__name__)

    class_name = re.sub("Schema", "", class_name)

    class_name = class_name.lower()

    return class_name


def revision_factory_join_field(record_instance: BaseModel) -> str:
    class_name = str(record_instance.__class__.__name__)

    FIELD_MAP = {
        "StationSchema": "station_id",
        "FacilitySchema": "facility_id",
        "LocationSchema": "location_id",
    }

    if class_name in FIELD_MAP.keys():
        return FIELD_MAP[class_name]

    raise Exception("Could not map record %s", record_instance)


def revision_factory(
    record: BaseModel, field_names: Union[str, List[str]], created_by: str,
) -> Optional[Revision]:

    if isinstance(field_names, str):
        field_names = [field_names]

    revision_data = {}

    for field_name in field_names:
        # field_type = get_schema_name(record)
        field_value = getattr(record, field_name)

        if isinstance(field_value, BaseModel):
            _value_dict = field_value.dict()

            # we only store the primary keys in associations rather
            # than the full record.
            if "id" in _value_dict:
                field_value = _value_dict["id"]

            elif "code" in _value_dict:
                field_value = _value_dict["code"]

            else:
                logger.error("Could not serialize data value %s", field_value)
                field_value = False

        if field_value is not False:
            revision_data[field_name] = field_value

    if not record.id:
        logger.error("Require a code to create a revision: %s", record)
        return None

    if len(revision_data.keys()) < 1:
        logger.error("Require data to lookup for code %s", record.code)
        return None

    column_name = revision_factory_join_field(record)

    __query = (
        s.query(Revision).filter(getattr(Revision, column_name) == record.id)
        # .filter(Revision.code == record.code)
    )

    for data_name, data_value in revision_data.items():
        if isinstance(data_value, str):
            __query = __query.filter(
                Revision.changes[data_name].as_string() == data_value
            )
        if isinstance(data_value, bool):
            __query = __query.filter(
                Revision.changes[data_name].as_boolean() == data_value
            )
        if isinstance(data_value, int):
            __query = __query.filter(
                Revision.changes[data_name].as_integer() == data_value
            )
        if isinstance(data_value, float):
            __query = __query.filter(
                Revision.changes[data_name].as_float() == data_value
            )

    revision_lookup = None

    try:
        revision_lookup = __query.one_or_none()
    except MultipleResultsFound:
        logger.info(
            "Revision exists: %s %s %s", record.code, field_name, field_value,
        )
        return None

    if revision_lookup:
        return None

    revision = Revision(created_by=created_by, changes=revision_data,)

    # s.add(revision)
    # s.commit()

    return revision


def load_revision(records, created_by):
    logger.info("Running db test")

    for station_record in records:
        station_model = (
            s.query(Station)
            .filter(Station.code == station_record.code)
            .one_or_none()
        )

        if not station_model:
            logger.info(
                f"New station {station_record.name} {station_record.code}"
            )

            station_dict = station_record.dict(
                include={"code", "network_name", "name", "location"}
            )
            # pprint(station_dict)

            # pylint:disable no-member
            station_model = fromdict(Station(), station_dict)
            station_model.approved = False
            station_model.created_by = created_by

            # location
            if "location" in station_dict and station_dict["location"]:
                station_model.location = fromdict(
                    Location(), station_dict["location"], exclude=["id"]
                )

                if station_record.location.lat and station_record.location.lng:
                    station_model.location.geom = station_record.location.geom

            s.add(station_model)
            s.commit()

            station_record.id = station_model.id

            # revision = revision_factory(
            #     station_record, ["code", "name", "network_name"], created_by,
            # )

            # if revision:
            #     station_model.revisions.append(revision)

            # s.add(station_model)
            # s.commit()

        else:
            for field in ["name"]:
                if getattr(station_model, field) != getattr(
                    station_record, field
                ):
                    revision_factory(station_record, field, created_by)

        for facility in station_record.facilities:
            facility_model = (
                s.query(Facility)
                .filter(Facility.code == facility.code)
                .first()
            )

            if not facility_model:
                logger.info(
                    "New facility %s => %s", station_record.name, facility.code
                )

                facility_dict = facility.dict(
                    include={
                        "code",
                        "network",
                        # "network_id",
                        "dispatch_type",
                        "station",
                        # "station_id",
                        # "status",
                        "network_code",
                        "network_region",
                        "network_name",
                    }
                )
                # pprint(station_dict)

                # pylint:disable no-member
                facility_model = fromdict(Facility(), facility_dict)
                facility_model.network_id = facility.network.code
                facility_model.approved = False
                facility_model.created_by = created_by

                s.add(facility_model)
                s.commit()

                facility.id = facility_model.id

                # @NOTE don't create revisions for new facilities
                # revision = revision_factory(
                #     facility, ["code", "dispatch_type"], created_by
                # )

                # if revision:
                #     facility_model.revisions.append(revision)

                # s.add(facility_model)
                # s.commit()

            else:
                facility.id = facility_model.id

            for field in [
                "fueltech",
                "status",
                "capacity_registered",
            ]:
                revision = None

                if compare_record_differs(facility_model, facility, field):
                    # logger.info(
                    #     "%s and %s differ",
                    #     getattr(facility, field),
                    #     getattr(facility_model, field),
                    # )

                    revision = revision_factory(facility, field, created_by)

                if revision:
                    facility_model.revisions.append(revision)

            s.add(facility_model)

            station_model.facilities.append(facility_model)
            s.add(station_model)
            s.commit()


def db_test():
    s.query(Revision).delete()

    mms = mms_import()
    rel = rel_import()
    gi = gi_import()

    load_revision(mms, "aemo.mms.202006")
    load_revision(rel, "aemo.rel.2020006")
    load_revision(gi, "aemo.gi.202006")


def registry_init():
    registry = registry_import()

    for station in registry:
        station_dict = station.dict(exclude={"id"})
        # pprint(station_dict)

        # pylint:disable no-member
        station_model = fromdict(Station(), station_dict)
        station_model.approved = True
        station_model.approved_at = datetime.now()
        station_model.approved_by = "opennem.registry"
        station_model.created_by = "opennem.registry"

        # location
        station_model.location = fromdict(
            Location(), station_dict["location"], exclude=["id"],
        )

        if station.location.lat and station.location.lng:
            station_model.location.geom = "SRID=4326;POINT({} {})".format(
                station.location.lng, station.location.lat
            )

        for fac in station.facilities:
            f = Facility(
                **fac.dict(
                    exclude={
                        "id",
                        "fueltech",
                        "status",
                        "network",
                        "revision_ids",
                        "scada_power",
                    }
                )
            )

            f.network_id = fac.network.code

            if fac.fueltech:
                f.fueltech_id = fac.fueltech.code

            f.status_id = fac.status.code

            f.approved = True
            f.approved_by = "opennem.registry"
            f.created_by = "opennem.registry"
            f.approved_at = datetime.now()
            f.created_at = datetime.now()

            station_model.facilities.append(f)

        s.add(station_model)

    s.commit()


def test_revisions():
    registry = registry_import()
    # mms = mms_import()

    k = registry.get_code("KWINANA")

    station = Station(code=k.code, name=k.name, network_name=k.network_name,)
    s.add(station)
    s.commit()

    # pylint: disable=no-member
    station_clone = station.asdict()
    station_clone.pop("id")

    pprint(station_clone)

    station = create_revision(station)

    station.location = Location(
        address1=k.location.address1,
        address2=k.location.address2,
        locality=k.location.locality,
        state=k.location.state,
        postcode=k.location.postcode,
    )

    s.add(station)
    s.commit()

    station = create_revision(station)

    station.name = "Kwinana Edited"

    s.add(station)
    s.commit()


def init():
    init_opennem(engine)
    # logger.info("Db initialized")

    load_fixtures()
    # logger.info("Fixtures loaded")

    registry_init()
    logger.info("Registry initialized")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    logger.setLevel(logging.INFO)

    # init()
    # db_test()
    registry_init()
    logger.info("Registry initialized")



import logging
from datetime import date, datetime, timedelta, timezone
from typing import List, Optional, Union

from dateutil.parser import ParserError, parse

from opennem.core.networks import NetworkSchema
from opennem.utils.timezone import UTC, is_aware, make_aware

logger = logging.getLogger(__name__)


def parse_date(
    date_str: Union[str, datetime],
    network: Optional[NetworkSchema] = None,
    dayfirst: bool = True,
    yearfirst: bool = False,
    is_utc: bool = False,
    timezone: timezone = False,
) -> Optional[datetime]:
    dt_return = None

    if isinstance(date_str, datetime):
        dt_return = date_str

    elif isinstance(date_str, str):
        try:
            dt_return = parse(date_str, dayfirst=dayfirst, yearfirst=yearfirst)
        except ParserError:
            raise ValueError("Invalid date string passed")

    else:
        raise ValueError("Require a datetime or string object to parse date")

    if network:
        tz = network.get_timezone()

        if tz:
            if is_aware(dt_return):
                if hasattr(tz, "localize"):
                    dt_return = tz.localize()
                else:
                    dt_return = dt_return.replace(tzinfo=tz)
            else:
                dt_return = make_aware(dt_return, timezone=tz)

    if is_utc:
        tz = UTC

        if is_aware(dt_return):
            if hasattr(tz, "localize"):
                dt_return = tz.localize()
            else:
                dt_return = dt_return.replace(tzinfo=tz)
        else:
            dt_return = make_aware(dt_return, timezone=tz)

    if timezone:
        dt_return = make_aware(dt_return, timezone=timezone)

    return dt_return


def get_date_component(format_str: str) -> str:
    """
        Get a part of a date
    """

    return datetime.now().strftime(format_str)


def date_series(
    start: Union[datetime, date] = None,
    end: Union[datetime, date] = None,
    length: int = 30,
    interval: timedelta = timedelta(days=1),
    reverse: bool = False,
) -> List[datetime]:
    """
        Generate a datetime series

    """
    if start and isinstance(start, datetime):
        start = start.date()

    if end and isinstance(end, datetime):
        end = start.date()

    if not start:
        start = datetime.now().date()

    if end:
        length = int(abs((end - start) / interval))

    next_record = start

    for _ in range(length):
        if reverse:
            next_record -= interval
        else:
            next_record += interval

        yield next_record



import json
import logging
import os
from datetime import date, datetime
from decimal import Decimal

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import MultipleResultsFound

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.loader import load_data
from opennem.core.normalizers import (
    clean_capacity,
    name_normalizer,
    normalize_duid,
    station_name_cleaner,
)
from opennem.core.station_duid_map import (
    facility_has_station_remap,
    facility_map_station,
)
from opennem.db import db_connect
from opennem.db.models.opennem import Facility, Station
from opennem.importer.compat import (
    map_compat_facility_state,
    map_compat_fueltech,
    map_compat_network_region,
)

logger = logging.getLogger(__name__)


def load_opennem_facilities():
    station_fixture = load_data("facility_registry.json", from_fixture=True)

    stations = [{"station_code": k, **v} for k, v in station_fixture.items()]

    engine = db_connect()
    session = sessionmaker(bind=engine)
    s = session()

    for station_data in stations:
        station = None

        facilities = [
            {"code": k, **v} for k, v in station_data["duid_data"].items()
        ]

        # update facilities
        for facility_data in facilities:
            facility_duid = facility_data["code"]
            station_name = station_name_cleaner(station_data["display_name"])
            # station_code = normalize_duid(station_data["station_code"])
            station_code = facility_map_station(
                facility_duid, normalize_duid(station_data["station_code"])
            )
            station_state = map_compat_facility_state(
                station_data["status"]["state"]
            )
            station_network = (
                "WEM" if station_data["location"]["state"] == "WA" else "NEM"
            )

            station = (
                s.query(Station)
                .filter(Station.network_code == station_code)
                .one_or_none()
            )

            if not station:
                station = Station(
                    network_id=station_network,
                    code=station_code,
                    network_code=station_code,
                    name=station_name,
                    network_name=station_data["display_name"],
                    created_by="opennem.load_facilities",
                )
                logger.info(
                    "Created station: {} {} ".format(
                        station_name, station_code
                    )
                )

                s.add(station)
                s.commit()

            facility_status = station_state

            # Network region
            facility_network_region = map_compat_network_region(
                station_data["region_id"]
            )

            # Fueltech
            facility_fueltech = None

            if "fuel_tech" in facility_data and facility_data["fuel_tech"]:
                facility_fueltech = map_compat_fueltech(
                    facility_data["fuel_tech"]
                )

            # Capacity
            facility_capacity = None

            if (
                "registered_capacity" in facility_data
                and facility_data["registered_capacity"]
            ):
                facility_capacity = clean_capacity(
                    facility_data["registered_capacity"]
                )

            facility = None

            try:
                facility = (
                    s.query(Facility)
                    .filter(Facility.network_code == facility_duid)
                    .one_or_none()
                )
            except MultipleResultsFound:
                logger.error(
                    "Multiple facilities found for duid {}".format(
                        facility_duid
                    )
                )

                # facility = (
                #     s.query(Facility)
                #     .filter(Facility.network_code == facility_duid)
                #     .first()
                # )
                continue

            if not facility:
                facility = Facility(
                    code=facility_duid,
                    network_code=facility_duid,
                    network_region=facility_network_region,
                    created_by="opennem.load_facilities",
                )

                logger.info(
                    "Created facility: {} {} to station {} ".format(
                        facility_duid, facility_fueltech, station_code
                    )
                )

            if not facility.unit_id:
                facility.unit_id = 1
                facility.unit_number = 1

            if facility_capacity and not facility.unit_capacity:
                facility.unit_capacity = facility_capacity

            if facility_capacity and not facility.capacity_registered:
                facility.capacity_registered = facility_capacity

            if facility_fueltech and not facility.fueltech_id:
                facility.fueltech_id = facility_fueltech

            if not facility.status_id:
                facility.status_id = facility_status

            if not facility.station:
                facility.station = station

            s.add(facility)
            s.commit()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    load_opennem_facilities()



# import pkg_resources
import sentry_sdk

from opennem.settings import settings

# VERSION = pkg_resources.get_distribution("opennem").version


if settings.env != "development":
    sentry_sdk.init(
        settings.sentry_url, traces_sample_rate=1.0, environment=settings.env
    )

# __version__ = VERSION



from opennem.db import db_connect
from opennem.db.models.opennem import metadata as metadata_opennem


def init_opennem(engine):
    metadata_opennem.drop_all(engine)
    metadata_opennem.create_all(engine)


if __name__ == "__main__":
    engine = db_connect()

    init_opennem(engine)



from typing import Optional

from pydantic import Field

from .core import BaseConfig


class UnitDefinition(BaseConfig):
    name: str = Field(..., description="Name of the unit")
    name_alias: Optional[str] = Field(None, description="Name alias")
    unit_type: str = Field(..., description="Type of unit")
    round_to: int = 2
    unit: str = Field(..., description="Unit abbreviation")




from typing import List

from fastapi import APIRouter, Depends, HTTPException
from starlette import status

from opennem.core.networks import NetworkNEM, NetworkWEM
from opennem.db import get_database_engine

from .schema import ScraperStats, ScraperStatsResult

router = APIRouter()


@router.get("/scraper/stats", response_model=List[ScraperStatsResult])
def scraper_stats(
    engine=Depends(get_database_engine),
) -> List[ScraperStatsResult]:
    query = """
        select
            count(fs.trading_interval),
            max(fs.trading_interval),
            min(fs.trading_interval)
        from facility_scada fs where fs.network_id='WEM'
        union select
            count(fs.trading_interval),
            max(fs.trading_interval),
            min(fs.trading_interval)
        from facility_scada fs where fs.network_id='NEM'
    """

    results = []

    with engine.connect() as c:
        results = list(c.execute(query))

    if not len(results):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Not implemented"
        )

    result = []

    result.append(
        ScraperStatsResult(
            network=NetworkWEM,
            stats=ScraperStats(
                scada_intervals=results[0][0],
                scada_max=results[0][1],
                scada_min=results[0][2],
            ),
        )
    )

    result.append(
        ScraperStatsResult(
            network=NetworkNEM,
            stats=ScraperStats(
                scada_intervals=results[1][0],
                scada_max=results[1][1],
                scada_min=results[1][2],
            ),
        )
    )

    return result



from sqlalchemy import (
    CHAR,
    DECIMAL,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Index,
    String,
    Table,
    Time,
    text,
)
from sqlalchemy.dialects.mysql import (
    BIGINT,
    INTEGER,
    MEDIUMINT,
    SMALLINT,
    TINYINT,
    VARCHAR,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()
metadata = Base.metadata


class BIDTYPE(Base):
    __tablename__ = "BIDTYPE"

    ID = Column(TINYINT(4), primary_key=True)
    BIDTYPE = Column(String(10))


t_DUID = Table(
    "DUID",
    metadata,
    Column("id", SMALLINT(6), server_default=text("'0'")),
    Column("DUID", String(10)),
    Column("STATION_NAME_ID", SMALLINT(6)),
    Column("PARTICIPANT_ID", SMALLINT(6)),
    Column("REGIONID", TINYINT(4)),
    Column("FUEL_SOURCE_ID", TINYINT(4)),
    Column("TECHNOLOGY_ID", TINYINT(4)),
    Column("CLASSIFICATION_ID", TINYINT(4)),
    Column("TECHFUEL_ID", TINYINT(4), server_default=text("'0'")),
    Column("MW", DECIMAL(29, 3)),
)


class DispatchedMarket(Base):
    __tablename__ = "DispatchedMarket"

    ID = Column(TINYINT(4), primary_key=True)
    DispatchedMarket = Column(String(30))


t_INTERCONNECTORID = Table(
    "INTERCONNECTORID",
    metadata,
    Column("id", TINYINT(4), server_default=text("'0'")),
    Column("INTERCONNECTORID", String(9)),
)


class INTERVALTABLE(Base):
    __tablename__ = "INTERVAL_TABLE"

    DISPATCHINTERVAL = Column(BIGINT(20), primary_key=True)
    TRADINGINTERVAL = Column(INTEGER(11), nullable=False, index=True)
    DATETIME = Column(DateTime, nullable=False, index=True)
    DATE = Column(Date, nullable=False, index=True)
    TIME = Column(Time, nullable=False, index=True)
    DISPATCHPERIOD = Column(SMALLINT(6), nullable=False, index=True)
    TRADINGPERIOD = Column(TINYINT(4), nullable=False, index=True)
    YEAR = Column(SMALLINT(6), nullable=False, index=True)
    MONTH = Column(TINYINT(4), nullable=False, index=True)
    WEEKDAY = Column(TINYINT(4), nullable=False, index=True)
    FY = Column(TINYINT(4), nullable=False, index=True)
    PEAK = Column(TINYINT(4), nullable=False)


class Market(Base):
    __tablename__ = "Market"

    ID = Column(TINYINT(4), primary_key=True)
    Market = Column(String(8))


t_REGIONID = Table(
    "REGIONID",
    metadata,
    Column("id", TINYINT(4), server_default=text("'0'")),
    Column("REGIONID", String(6)),
)


class RUNTYPE(Base):
    __tablename__ = "RUNTYPE"

    ID = Column(TINYINT(4), primary_key=True)
    RUNTYPE = Column(String(45), unique=True)


class Unit(Base):
    __tablename__ = "Unit"

    ID = Column(SMALLINT(6), primary_key=True)
    Unit = Column(VARCHAR(30), unique=True)


class DUID(Base):
    __tablename__ = "DUID"
    __table_args__ = {"schema": "nemweb_meta"}

    ID = Column(SMALLINT(6), primary_key=True)
    DUID = Column(VARCHAR(10), unique=True)


class REGIONID(Base):
    __tablename__ = "REGIONID"
    __table_args__ = {"schema": "nemweb_meta"}

    id = Column(TINYINT(4), primary_key=True)
    REGIONID = Column(VARCHAR(6), unique=True)


class CO2EIIPUBLISHING(Base):
    __tablename__ = "CO2EII_PUBLISHING"
    __table_args__ = (
        Index("SETTLEMENTDATE_2", "SETTLEMENTDATE", "REGIONID", unique=True),
    )

    id = Column(INTEGER(11), primary_key=True)
    CONTRACTYEAR = Column(SMALLINT(6), index=True)
    WEEKNO = Column(TINYINT(4), index=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    TOTAL_SENT_OUT_ENERGY = Column(DECIMAL(14, 8))
    TOTAL_EMISSIONS = Column(DECIMAL(14, 8))
    CO2E_INTENSITY_INDEX = Column(DECIMAL(5, 4))

    REGIONID1 = relationship("REGIONID")


class DISPATCHPRICE(Base):
    __tablename__ = "DISPATCH_PRICE"
    __table_args__ = (
        Index(
            "uniq", "SETTLEMENTDATE", "REGIONID", "INTERVENTION", unique=True
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DISPATCHINTERVAL = Column(BIGINT(20), index=True)
    RRP = Column(DECIMAL(10, 5))
    ROP = Column(DECIMAL(12, 5))
    APCFLAG = Column(TINYINT(4), index=True)
    CUMUL_PRE_AP_ENERGY_PRICE = Column(DECIMAL(16, 10))
    INTERVENTION = Column(
        TINYINT(4), nullable=False, index=True, server_default=text("0")
    )
    RAISE6SECRRP = Column(DECIMAL(12, 5))
    RAISE6SECROP = Column(DECIMAL(12, 5))
    RAISE6SECAPCFLAG = Column(TINYINT(4), index=True)
    RAISE60SECRRP = Column(DECIMAL(12, 5))
    RAISE60SECROP = Column(DECIMAL(12, 5))
    RAISE60SECAPCFLAG = Column(TINYINT(4), index=True)
    RAISE5MINRRP = Column(DECIMAL(12, 5))
    RAISE5MINROP = Column(DECIMAL(12, 5))
    RAISE5MINAPCFLAG = Column(TINYINT(4), index=True)
    RAISEREGRRP = Column(DECIMAL(12, 5))
    RAISEREGROP = Column(DECIMAL(12, 5))
    RAISEREGAPCFLAG = Column(TINYINT(4), index=True)
    LOWER6SECRRP = Column(DECIMAL(12, 5))
    LOWER6SECROP = Column(DECIMAL(12, 5))
    LOWER6SECAPCFLAG = Column(TINYINT(4), index=True)
    LOWER60SECRRP = Column(DECIMAL(12, 5))
    LOWER60SECROP = Column(DECIMAL(12, 5))
    LOWER60SECAPCFLAG = Column(TINYINT(4), index=True)
    LOWER5MINRRP = Column(DECIMAL(12, 5))
    LOWER5MINROP = Column(DECIMAL(12, 5))
    LOWER5MINAPCFLAG = Column(TINYINT(4), index=True)
    LOWERREGRRP = Column(DECIMAL(12, 5))
    LOWERREGROP = Column(DECIMAL(12, 5))
    LOWERREGAPCFLAG = Column(TINYINT(4), index=True)

    REGIONID1 = relationship("REGIONID")


class DISPATCHREGIONSUM(Base):
    __tablename__ = "DISPATCH_REGIONSUM"
    __table_args__ = (
        Index(
            "uniq", "REGIONID", "SETTLEMENTDATE", "INTERVENTION", unique=True
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    TOTALDEMAND = Column(DECIMAL(7, 2))
    AVAILABLEGENERATION = Column(DECIMAL(10, 5))
    AVAILABLELOAD = Column(DECIMAL(7, 2))
    DEMANDFORECAST = Column(DECIMAL(10, 5))
    DISPATCHABLEGENERATION = Column(DECIMAL(7, 2))
    DISPATCHABLELOAD = Column(DECIMAL(7, 2))
    NETINTERCHANGE = Column(DECIMAL(6, 2))
    EXCESSGENERATION = Column(DECIMAL(7, 2))
    INITIALSUPPLY = Column(DECIMAL(10, 5))
    CLEAREDSUPPLY = Column(DECIMAL(7, 2))
    TOTALINTERMITTENTGENERATION = Column(DECIMAL(10, 5))
    DEMAND_AND_NONSCHEDGEN = Column(DECIMAL(10, 5))
    UIGF = Column(DECIMAL(10, 5))
    SEMISCHEDULE_CLEAREDMW = Column(DECIMAL(10, 5))
    DISPATCHINTERVAL = Column(BIGINT(20), index=True)
    SEMISCHEDULE_COMPLIANCEMW = Column(DECIMAL(10, 5))
    INTERVENTION = Column(
        TINYINT(4), nullable=False, index=True, server_default=text("0")
    )
    LOWERREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER6SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISEREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE6SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    AGGREGATEDISPATCHERROR = Column(DECIMAL(10, 5))
    RAISE6SECACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    RAISE5MINACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    RAISE60SECACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    RAISEREGACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    LOWER6SECACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    LOWER5MINACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    LOWER60SECACTUALAVAILABILITY = Column(DECIMAL(12, 6))
    LOWERREGACTUALAVAILABILITY = Column(DECIMAL(12, 6))

    REGIONID1 = relationship("REGIONID")


class DISPATCHUNITSCADA(Base):
    __tablename__ = "DISPATCH_UNIT_SCADA"
    __table_args__ = (Index("uniq", "SETTLEMENTDATE", "DUID", unique=True),)

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DUID = Column(
        ForeignKey("nemweb_meta.DUID.ID"),
        ForeignKey("nemweb_meta.DUID.ID"),
        index=True,
    )
    SCADAVALUE = Column(DECIMAL(10, 6))

    DUID1 = relationship(
        "DUID", primaryjoin="DISPATCHUNITSCADA.DUID == DUID.ID"
    )
    DUID2 = relationship(
        "DUID", primaryjoin="DISPATCHUNITSCADA.DUID == DUID.ID"
    )


class DISPATCHUNITSOLUTION(Base):
    __tablename__ = "DISPATCH_UNIT_SOLUTION"
    __table_args__ = (
        Index(
            "uniq_idx", "DUID", "SETTLEMENTDATE", "INTERVENTION", unique=True
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    INTERVENTION = Column(TINYINT(4), index=True)
    DISPATCHMODE = Column(TINYINT(4), index=True)
    AGCSTATUS = Column(TINYINT(4), index=True)
    INITIALMW = Column(DECIMAL(10, 6))
    TOTALCLEARED = Column(DECIMAL(10, 6))
    RAMPDOWNRATE = Column(DECIMAL(8, 2))
    RAMPUPRATE = Column(DECIMAL(8, 2))
    LOWER5MIN = Column(DECIMAL(10, 6))
    LOWER60SEC = Column(DECIMAL(10, 6))
    LOWER6SEC = Column(DECIMAL(10, 6))
    RAISE5MIN = Column(DECIMAL(10, 6))
    RAISE60SEC = Column(DECIMAL(10, 6))
    RAISE6SEC = Column(DECIMAL(10, 6))
    LOWERREG = Column(DECIMAL(10, 6))
    RAISEREG = Column(DECIMAL(10, 6))
    AVAILABILITY = Column(DECIMAL(10, 6))

    DUID1 = relationship("DUID")


class DISPATCHUNITSOLUTIONOLD(Base):
    __tablename__ = "DISPATCH_UNIT_SOLUTION_OLD"
    __table_args__ = (
        Index("uniq", "SETTLEMENTDATE", "DUID", "INTERVENTION", unique=True),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    INITIALMW = Column(DECIMAL(10, 6))
    TOTALCLEARED = Column(DECIMAL(10, 6))
    INTERVENTION = Column(SMALLINT(6), index=True)

    DUID1 = relationship("DUID")


class METERDATAGENDUID(Base):
    __tablename__ = "METER_DATA_GEN_DUID"
    __table_args__ = (
        Index("DUID_2", "DUID", "INTERVAL_DATETIME", unique=True),
    )

    ID = Column(INTEGER(11), primary_key=True)
    INTERVAL_DATETIME = Column(DateTime, index=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    MWH_READING = Column(DECIMAL(10, 6))

    DUID1 = relationship("DUID")


class MTPASAREGIONSOLUTION(Base):
    __tablename__ = "MTPASA_REGIONSOLUTION"
    __table_args__ = (
        Index(
            "uniq",
            "RUN_DATETIME",
            "RUN_NO",
            "DAY",
            "REGIONID",
            "RUNTYPE",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    RUN_DATETIME = Column(DateTime, index=True)
    RUN_NO = Column(INTEGER(11), index=True)
    RUNTYPE = Column(ForeignKey("RUNTYPE.ID"), index=True)
    DAY = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    DEMAND10 = Column(DECIMAL(12, 2))
    ENERGYREQDEMAND10 = Column(DECIMAL(12, 2))
    UNCONSTRAINEDCAPACITY = Column(DECIMAL(12, 0))
    CONSTRAINEDCAPACITY = Column(DECIMAL(12, 0))
    NETINTERCHANGEUNDERSCARCITY = Column(DECIMAL(12, 2))
    SURPLUSCAPACITY = Column(DECIMAL(12, 2))
    RESERVECONDITION = Column(TINYINT(4))
    MAXSURPLUSRESERVE = Column(DECIMAL(12, 2))
    MAXSPARECAPACITY = Column(DECIMAL(12, 2))
    LORCONDITION = Column(TINYINT(4))
    AGGREGATECAPACITYAVAILABLE = Column(DECIMAL(12, 2))
    AGGREGATESCHEDULEDLOAD = Column(DECIMAL(12, 2))
    AGGREGATEPASAAVAILABILITY = Column(DECIMAL(12, 0))
    TOTALINTERMITTENTGENERATION = Column(DECIMAL(10, 5))
    DEMAND50 = Column(DECIMAL(12, 2))
    DEMAND_AND_NONSCHEDGEN = Column(DECIMAL(15, 5))
    UIGF = Column(DECIMAL(12, 2))
    SEMISCHEDULEDCAPACITY = Column(DECIMAL(12, 2))
    DEFICITRESERVE = Column(DECIMAL(16, 6))
    ENERGYREQDEMAND50 = Column(DECIMAL(12, 2))

    REGIONID1 = relationship("REGIONID")
    RUNTYPE1 = relationship("RUNTYPE")


class NEMPriceSetter(Base):
    __tablename__ = "NEMPriceSetter"
    __table_args__ = (
        Index(
            "uniq_idx",
            "Market",
            "Unit",
            "Unit_2",
            "PeriodID",
            "REGIONID",
            "DispatchedMarket",
            "OCD_flag",
            "BandNo",
            "BandNo_2",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    PeriodID = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    Market = Column(ForeignKey("Market.ID"), index=True)
    DispatchedMarket = Column(ForeignKey("DispatchedMarket.ID"), index=True)
    Price = Column(DECIMAL(12, 5))
    Unit = Column(ForeignKey("Unit.ID"), index=True)
    BandNo = Column(TINYINT(4))
    Unit_2 = Column(ForeignKey("Unit.ID"), index=True)
    BandNo_2 = Column(TINYINT(4))
    Increase = Column(DECIMAL(12, 5))
    RRNBandPrice = Column(DECIMAL(10, 2))
    BandCost = Column(DECIMAL(14, 6))
    OCD_flag = Column(TINYINT(4), index=True, server_default=text("0"))

    DispatchedMarket1 = relationship("DispatchedMarket")
    Market1 = relationship("Market")
    REGIONID1 = relationship("REGIONID")
    Unit1 = relationship("Unit", primaryjoin="NEMPriceSetter.Unit == Unit.ID")
    Unit2 = relationship(
        "Unit", primaryjoin="NEMPriceSetter.Unit_2 == Unit.ID"
    )


class OFFERBIDDAYOFFER(Base):
    __tablename__ = "OFFER_BIDDAYOFFER"

    ID = Column(INTEGER(11), primary_key=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    OFFERDATE = Column(DateTime, index=True)
    VERSIONNO = Column(SMALLINT(6), index=True)
    BIDTYPE = Column(ForeignKey("BIDTYPE.ID"), index=True)
    DAILYENERGYCONSTRAINT = Column(INTEGER(11))
    PRICEBAND1 = Column(DECIMAL(7, 2))
    PRICEBAND2 = Column(DECIMAL(7, 2))
    PRICEBAND3 = Column(DECIMAL(7, 2))
    PRICEBAND4 = Column(DECIMAL(7, 2))
    PRICEBAND5 = Column(DECIMAL(7, 2))
    PRICEBAND6 = Column(DECIMAL(7, 2))
    PRICEBAND7 = Column(DECIMAL(7, 2))
    PRICEBAND8 = Column(DECIMAL(7, 2))
    PRICEBAND9 = Column(DECIMAL(7, 2))
    PRICEBAND10 = Column(DECIMAL(7, 2))
    MINIMUMLOAD = Column(MEDIUMINT(9))
    LASTCHANGED = Column(DateTime, index=True)
    ENTRYTYPE = Column(CHAR(5), index=True)
    T1 = Column(TINYINT(4))
    T2 = Column(TINYINT(4))
    T3 = Column(TINYINT(4))
    T4 = Column(TINYINT(4))

    BIDTYPE1 = relationship("BIDTYPE")
    DUID1 = relationship("DUID")


class OFFERBIDPEROFFER(Base):
    __tablename__ = "OFFER_BIDPEROFFER"

    ID = Column(INTEGER(11), primary_key=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    OFFERDATE = Column(DateTime, index=True)
    PERIODID = Column(TINYINT(4), index=True)
    VERSIONNO = Column(MEDIUMINT(9), index=True)
    BIDTYPE = Column(ForeignKey("BIDTYPE.ID"), index=True)
    MAXAVAIL = Column(MEDIUMINT(9))
    ROCUP = Column(SMALLINT(6))
    ROCDOWN = Column(SMALLINT(6))
    BANDAVAIL1 = Column(SMALLINT(6))
    BANDAVAIL2 = Column(SMALLINT(6))
    BANDAVAIL3 = Column(SMALLINT(6))
    BANDAVAIL4 = Column(SMALLINT(6))
    BANDAVAIL5 = Column(SMALLINT(6))
    BANDAVAIL6 = Column(SMALLINT(6))
    BANDAVAIL7 = Column(SMALLINT(6))
    BANDAVAIL8 = Column(SMALLINT(6))
    BANDAVAIL9 = Column(SMALLINT(6))
    BANDAVAIL10 = Column(SMALLINT(6))
    LASTCHANGED = Column(DateTime, index=True)

    BIDTYPE1 = relationship("BIDTYPE")
    DUID1 = relationship("DUID")


class P5MINREGIONSOLUTION(Base):
    __tablename__ = "P5MIN_REGIONSOLUTION"
    __table_args__ = (
        Index(
            "uniq",
            "RUN_DATETIME",
            "INTERVAL_DATETIME",
            "REGIONID",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    RUN_DATETIME = Column(DateTime, index=True)
    INTERVAL_DATETIME = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    RRP = Column(DECIMAL(10, 5))
    EXCESSGENERATION = Column(DECIMAL(7, 2))
    RAISE6SECRRP = Column(DECIMAL(10, 5))
    RAISE60SECRRP = Column(DECIMAL(10, 5))
    RAISE5MINRRP = Column(DECIMAL(10, 5))
    RAISEREGRRP = Column(DECIMAL(10, 5))
    LOWER6SECRRP = Column(DECIMAL(10, 5))
    LOWER60SECRRP = Column(DECIMAL(10, 5))
    LOWER5MINRRP = Column(DECIMAL(10, 5))
    LOWERREGRRP = Column(DECIMAL(10, 5))
    TOTALDEMAND = Column(DECIMAL(7, 2))
    AVAILABLEGENERATION = Column(DECIMAL(8, 3))
    AVAILABLELOAD = Column(DECIMAL(7, 3))
    DEMANDFORECAST = Column(DECIMAL(10, 5))
    DISPATCHABLEGENERATION = Column(DECIMAL(7, 2))
    DISPATCHABLELOAD = Column(DECIMAL(7, 2))
    NETINTERCHANGE = Column(DECIMAL(7, 2))
    LOWER5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER6SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE6SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    AGGREGATEDISPATCHERROR = Column(DECIMAL(10, 5))
    INITIALSUPPLY = Column(DECIMAL(10, 5))
    CLEAREDSUPPLY = Column(DECIMAL(7, 2))
    LOWERREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISEREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    TOTALINTERMITTENTGENERATION = Column(DECIMAL(8, 3))
    DEMAND_AND_NONSCHEDGEN = Column(DECIMAL(8, 3))
    UIGF = Column(DECIMAL(8, 3))
    SEMISCHEDULE_CLEAREDMW = Column(DECIMAL(9, 4))
    SEMISCHEDULE_COMPLIANCEMW = Column(DECIMAL(9, 4))

    REGIONID1 = relationship("REGIONID")


class PREDISPATCHPRICESENSITIVITY(Base):
    __tablename__ = "PREDISPATCH_PRICESENSITIVITIES"
    __table_args__ = (
        Index(
            "uniq_idx",
            "PREDISPATCHSEQNO",
            "REGIONID",
            "PERIODID",
            "INTERVENTION",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    PREDISPATCHSEQNO = Column(BIGINT(20), index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    PERIODID = Column(TINYINT(4), index=True)
    INTERVENTION = Column(TINYINT(4), index=True)
    INTERVENTION_ACTIVE = Column(TINYINT(4), index=True)
    DATETIME = Column(DateTime, index=True)
    RRPEEP1 = Column(DECIMAL(12, 6))
    RRPEEP2 = Column(DECIMAL(12, 6))
    RRPEEP3 = Column(DECIMAL(12, 6))
    RRPEEP4 = Column(DECIMAL(12, 6))
    RRPEEP5 = Column(DECIMAL(12, 6))
    RRPEEP6 = Column(DECIMAL(12, 6))
    RRPEEP7 = Column(DECIMAL(12, 6))
    RRPEEP8 = Column(DECIMAL(12, 6))
    RRPEEP9 = Column(DECIMAL(12, 6))
    RRPEEP10 = Column(DECIMAL(12, 6))
    RRPEEP11 = Column(DECIMAL(12, 6))
    RRPEEP12 = Column(DECIMAL(12, 6))
    RRPEEP13 = Column(DECIMAL(12, 6))
    RRPEEP14 = Column(DECIMAL(12, 6))
    RRPEEP15 = Column(DECIMAL(12, 6))
    RRPEEP16 = Column(DECIMAL(12, 6))
    RRPEEP17 = Column(DECIMAL(12, 6))
    RRPEEP18 = Column(DECIMAL(12, 6))
    RRPEEP19 = Column(DECIMAL(12, 6))
    RRPEEP20 = Column(DECIMAL(12, 6))
    RRPEEP21 = Column(DECIMAL(12, 6))
    RRPEEP22 = Column(DECIMAL(12, 6))
    RRPEEP23 = Column(DECIMAL(12, 6))
    RRPEEP24 = Column(DECIMAL(12, 6))
    RRPEEP25 = Column(DECIMAL(12, 6))
    RRPEEP26 = Column(DECIMAL(12, 6))
    RRPEEP27 = Column(DECIMAL(12, 6))
    RRPEEP28 = Column(DECIMAL(12, 6))
    RRPEEP29 = Column(DECIMAL(12, 6))
    RRPEEP30 = Column(DECIMAL(12, 6))
    RRPEEP31 = Column(DECIMAL(12, 6))
    RRPEEP32 = Column(DECIMAL(12, 6))
    RRPEEP33 = Column(DECIMAL(12, 6))
    RRPEEP34 = Column(DECIMAL(12, 6))
    RRPEEP35 = Column(DECIMAL(12, 6))
    RRPEEP36 = Column(DECIMAL(12, 6))
    RRPEEP37 = Column(DECIMAL(12, 6))
    RRPEEP38 = Column(DECIMAL(12, 6))
    RRPEEP39 = Column(DECIMAL(12, 6))
    RRPEEP40 = Column(DECIMAL(12, 6))
    RRPEEP41 = Column(DECIMAL(12, 6))
    RRPEEP42 = Column(DECIMAL(12, 6))
    RRPEEP43 = Column(DECIMAL(12, 6))
    LASTCHANGED = Column(DateTime)

    REGIONID1 = relationship("REGIONID")


class PREDISPATCHREGIONPRICE(Base):
    __tablename__ = "PREDISPATCH_REGION_PRICES"
    __table_args__ = (
        Index(
            "uniq",
            "REGIONID",
            "PREDISPATCHSEQNO",
            "INTERVENTION",
            "PERIODID",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    PREDISPATCHSEQNO = Column(INTEGER(11), index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    PERIODID = Column(TINYINT(4), index=True)
    RRP = Column(DECIMAL(10, 5))
    DATETIME = Column(DateTime, index=True)
    INTERVENTION = Column(TINYINT(4), nullable=False, index=True)
    RAISE60SECRRP = Column(DECIMAL(10, 5))
    RAISE5MINRRP = Column(DECIMAL(10, 5))
    RAISEREGRRP = Column(DECIMAL(10, 5))
    RAISE6SECRRP = Column(DECIMAL(10, 5))
    LOWER6SECRRP = Column(DECIMAL(10, 5))
    LOWER60SECRRP = Column(DECIMAL(10, 5))
    LOWER5MINRRP = Column(DECIMAL(10, 5))
    LOWERREGRRP = Column(DECIMAL(10, 5))

    REGIONID1 = relationship("REGIONID")


class PREDISPATCHREGIONSOLUTION(Base):
    __tablename__ = "PREDISPATCH_REGION_SOLUTION"
    __table_args__ = (
        Index(
            "uniq",
            "REGIONID",
            "PREDISPATCHSEQNO",
            "INTERVENTION",
            "PERIODID",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    PREDISPATCHSEQNO = Column(INTEGER(11), index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    TOTALDEMAND = Column(DECIMAL(7, 2))
    AVAILABLEGENERATION = Column(DECIMAL(8, 3))
    AVAILABLELOAD = Column(DECIMAL(7, 3))
    DEMANDFORECAST = Column(DECIMAL(10, 5))
    DISPATCHABLEGENERATION = Column(DECIMAL(7, 2))
    DISPATCHABLELOAD = Column(DECIMAL(7, 2))
    NETINTERCHANGE = Column(DECIMAL(7, 2), nullable=False)
    EXCESSGENERATION = Column(DECIMAL(7, 2))
    DATETIME = Column(DateTime, index=True)
    INITIALSUPPLY = Column(DECIMAL(10, 5))
    CLEAREDSUPPLY = Column(DECIMAL(7, 2))
    TOTALINTERMITTENTGENERATION = Column(DECIMAL(8, 3))
    DEMAND_AND_NONSCHEDGEN = Column(DECIMAL(8, 3))
    UIGF = Column(DECIMAL(8, 3))
    SEMISCHEDULE_CLEAREDMW = Column(DECIMAL(9, 4), nullable=False)
    SEMISCHEDULE_COMPLIANCEMW = Column(DECIMAL(9, 4))
    PERIODID = Column(TINYINT(4), index=True)
    INTERVENTION = Column(TINYINT(4), nullable=False, index=True)

    REGIONID1 = relationship("REGIONID")


class ROOFTOPACTUAL(Base):
    __tablename__ = "ROOFTOP_ACTUAL"
    __table_args__ = (
        Index("uniq_idx", "INTERVAL_DATETIME", "REGIONID", unique=True),
    )

    ID = Column(INTEGER(11), primary_key=True)
    INTERVAL_DATETIME = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    POWER = Column(DECIMAL(8, 3))

    REGIONID1 = relationship("REGIONID")


class ROOFTOPFORECAST(Base):
    __tablename__ = "ROOFTOP_FORECAST"
    __table_args__ = (
        Index(
            "uniq_idx",
            "VERSION_DATETIME",
            "INTERVAL_DATETIME",
            "REGIONID",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    VERSION_DATETIME = Column(DateTime, index=True)
    INTERVAL_DATETIME = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    POWERMEAN = Column(DECIMAL(8, 3))
    POWERPOE50 = Column(DECIMAL(8, 3))
    POWERPOELOW = Column(DECIMAL(8, 3))
    POWERPOEHIGH = Column(DECIMAL(8, 3))

    REGIONID1 = relationship("REGIONID")


class TRADINGPRICE(Base):
    __tablename__ = "TRADING_PRICE"
    __table_args__ = (
        Index("uniq", "SETTLEMENTDATE", "REGIONID", unique=True),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, nullable=False, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    RRP = Column(DECIMAL(7, 2))
    ROP = Column(DECIMAL(8, 2))
    RAISE6SECRRP = Column(DECIMAL(12, 5))
    RAISE6SECROP = Column(DECIMAL(12, 5))
    RAISE60SECRRP = Column(DECIMAL(12, 5))
    RAISE60SECROP = Column(DECIMAL(12, 5))
    RAISE5MINRRP = Column(DECIMAL(12, 5))
    RAISE5MINROP = Column(DECIMAL(12, 5))
    RAISEREGRRP = Column(DECIMAL(12, 5))
    RAISEREGROP = Column(DECIMAL(12, 5))
    LOWER6SECRRP = Column(DECIMAL(12, 5))
    LOWER6SECROP = Column(DECIMAL(12, 5))
    LOWER60SECRRP = Column(DECIMAL(12, 5))
    LOWER60SECROP = Column(DECIMAL(12, 5))
    LOWER5MINRRP = Column(DECIMAL(12, 5))
    LOWER5MINROP = Column(DECIMAL(12, 5))
    LOWERREGRRP = Column(DECIMAL(12, 5))
    LOWERREGROP = Column(DECIMAL(12, 5))

    REGIONID1 = relationship("REGIONID")


class TRADINGREGIONSUM(Base):
    __tablename__ = "TRADING_REGIONSUM"
    __table_args__ = (
        Index("uniq", "SETTLEMENTDATE", "REGIONID", unique=True),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    TOTALDEMAND = Column(DECIMAL(7, 2))
    AVAILABLEGENERATION = Column(DECIMAL(7, 2))
    AVAILABLELOAD = Column(DECIMAL(7, 2))
    DEMANDFORECAST = Column(DECIMAL(7, 2))
    DISPATCHABLEGENERATION = Column(DECIMAL(7, 2))
    DISPATCHABLELOAD = Column(DECIMAL(7, 2))
    NETINTERCHANGE = Column(DECIMAL(7, 2))
    EXCESSGENERATION = Column(DECIMAL(7, 2))
    INITIALSUPPLY = Column(DECIMAL(7, 2))
    CLEAREDSUPPLY = Column(DECIMAL(7, 2))
    TOTALINTERMITTENTGENERATION = Column(DECIMAL(7, 2))
    DEMAND_AND_NONSCHEDGEN = Column(DECIMAL(7, 2))
    UIGF = Column(DECIMAL(7, 2))
    LOWERREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    LOWER6SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE5MINLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISEREGLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE60SECLOCALDISPATCH = Column(DECIMAL(10, 5))
    RAISE6SECLOCALDISPATCH = Column(DECIMAL(10, 5))

    REGIONID1 = relationship("REGIONID")


class YESTBIDBIDDAYOFFER(Base):
    __tablename__ = "YESTBID_BIDDAYOFFER"
    __table_args__ = (
        Index(
            "uniq",
            "DUID",
            "LASTCHANGED",
            "BIDOFFERDATE",
            "BIDTYPE",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    BIDOFFERDATE = Column(DateTime, index=True)
    PRICEBAND1 = Column(DECIMAL(7, 2))
    PRICEBAND2 = Column(DECIMAL(7, 2))
    PRICEBAND3 = Column(DECIMAL(7, 2))
    PRICEBAND4 = Column(DECIMAL(7, 2))
    PRICEBAND5 = Column(DECIMAL(7, 2), nullable=False)
    PRICEBAND6 = Column(DECIMAL(7, 2))
    PRICEBAND7 = Column(DECIMAL(7, 2))
    PRICEBAND8 = Column(DECIMAL(7, 2))
    PRICEBAND9 = Column(DECIMAL(7, 2), nullable=False)
    PRICEBAND10 = Column(DECIMAL(7, 2))
    MINIMUMLOAD = Column(SMALLINT(6))
    LASTCHANGED = Column(DateTime, index=True)
    BIDVERSIONNO = Column(SMALLINT(6))
    ENTRYTYPE = Column(String(10), index=True)
    BIDTYPE = Column(ForeignKey("BIDTYPE.ID"), index=True)
    BIDSETTLEMENTDATE = Column(DateTime)
    DAILYENERGYCONSTRAINT = Column(DECIMAL(12, 6))

    BIDTYPE1 = relationship("BIDTYPE")
    DUID1 = relationship("DUID")


class YESTBIDBIDPEROFFER(Base):
    __tablename__ = "YESTBID_BIDPEROFFER"
    __table_args__ = (
        Index(
            "uniq",
            "DUID",
            "TRADINGPERIOD",
            "LASTCHANGED",
            "BIDTYPE",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    DUID = Column(ForeignKey("nemweb_meta.DUID.ID"), index=True)
    BIDOFFERDATE = Column(DateTime, index=True)
    TRADINGPERIOD = Column(DateTime, index=True)
    MAXAVAIL = Column(SMALLINT(6))
    FIXEDLOAD = Column(SMALLINT(6))
    ROCUP = Column(SMALLINT(6))
    ROCDOWN = Column(SMALLINT(6))
    BANDAVAIL1 = Column(SMALLINT(6))
    BANDAVAIL2 = Column(SMALLINT(6))
    BANDAVAIL3 = Column(SMALLINT(6))
    BANDAVAIL4 = Column(SMALLINT(6))
    BANDAVAIL5 = Column(SMALLINT(6))
    BANDAVAIL6 = Column(SMALLINT(6))
    BANDAVAIL7 = Column(SMALLINT(6))
    BANDAVAIL8 = Column(SMALLINT(6))
    BANDAVAIL9 = Column(SMALLINT(6))
    BANDAVAIL10 = Column(SMALLINT(6))
    PASAAVAILABILITY = Column(SMALLINT(6))
    PERIODID = Column(TINYINT(4), nullable=False, index=True)
    LASTCHANGED = Column(DateTime, index=True)
    BIDVERSIONNO = Column(SMALLINT(6))
    SETTLEMENTDATE = Column(DateTime)
    BIDTYPE = Column(ForeignKey("BIDTYPE.ID"), index=True)
    BIDSETTLEMENTDATE = Column(DateTime)
    ENABLEMENTMIN = Column(DECIMAL(6, 0))
    ENABLEMENTMAX = Column(DECIMAL(6, 0))
    HIGHBREAKPOINT = Column(DECIMAL(6, 0))
    LOWBREAKPOINT = Column(DECIMAL(6, 0))

    BIDTYPE1 = relationship("BIDTYPE")
    DUID1 = relationship("DUID")


class INTERCONNECTORID(Base):
    __tablename__ = "INTERCONNECTORID"
    __table_args__ = {"schema": "nemweb_meta"}

    id = Column(TINYINT(4), primary_key=True)
    INTERCONNECTORID = Column(VARCHAR(9), nullable=False, unique=True)
    FROM_REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)
    TO_REGIONID = Column(ForeignKey("nemweb_meta.REGIONID.id"), index=True)

    REGIONID = relationship(
        "REGIONID", primaryjoin="INTERCONNECTORID.FROM_REGIONID == REGIONID.id"
    )
    REGIONID1 = relationship(
        "REGIONID", primaryjoin="INTERCONNECTORID.TO_REGIONID == REGIONID.id"
    )


class DISPATCHINTERCONNECTORRE(Base):
    __tablename__ = "DISPATCH_INTERCONNECTORRES"
    __table_args__ = (
        Index(
            "uniq",
            "INTERCONNECTORID",
            "SETTLEMENTDATE",
            "INTERVENTION",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    INTERCONNECTORID = Column(
        ForeignKey("nemweb_meta.INTERCONNECTORID.id"), index=True
    )
    METEREDMWFLOW = Column(DECIMAL(10, 5))
    MWLOSSES = Column(DECIMAL(8, 5))
    EXPORTLIMIT = Column(DECIMAL(10, 5))
    IMPORTLIMIT = Column(DECIMAL(10, 5))
    MARGINALLOSS = Column(DECIMAL(6, 5))
    DISPATCHINTERVAL = Column(BIGINT(20), index=True)
    MWFLOW = Column(DECIMAL(10, 5))
    INTERVENTION = Column(
        TINYINT(4), nullable=False, index=True, server_default=text("0")
    )

    INTERCONNECTORID1 = relationship("INTERCONNECTORID")


class MTPASAINTERCONNECTORSOLUTION(Base):
    __tablename__ = "MTPASA_INTERCONNECTORSOLUTION"
    __table_args__ = (
        Index(
            "uniq",
            "RUN_DATETIME",
            "RUN_NO",
            "DAY",
            "INTERCONNECTORID",
            "RUNTYPE",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    RUN_DATETIME = Column(DateTime, index=True)
    RUN_NO = Column(INTEGER(11), index=True)
    RUNTYPE = Column(ForeignKey("RUNTYPE.ID"), index=True)
    DAY = Column(DateTime, index=True)
    INTERCONNECTORID = Column(
        ForeignKey("nemweb_meta.INTERCONNECTORID.id"), index=True
    )
    CAPACITYMWFLOW = Column(DECIMAL(12, 2))
    CALCULATEDEXPORTLIMIT = Column(DECIMAL(12, 2))
    CALCULATEDIMPORTLIMIT = Column(DECIMAL(12, 2))

    INTERCONNECTORID1 = relationship("INTERCONNECTORID")
    RUNTYPE1 = relationship("RUNTYPE")


class P5MININTERCONNECTORSOLN(Base):
    __tablename__ = "P5MIN_INTERCONNECTORSOLN"
    __table_args__ = (
        Index(
            "uniq",
            "RUN_DATETIME",
            "INTERCONNECTORID",
            "INTERVAL_DATETIME",
            unique=True,
        ),
    )

    ID = Column(INTEGER(11), primary_key=True)
    RUN_DATETIME = Column(DateTime, index=True)
    INTERCONNECTORID = Column(
        ForeignKey("nemweb_meta.INTERCONNECTORID.id"), index=True
    )
    INTERVAL_DATETIME = Column(DateTime, index=True)
    METEREDMWFLOW = Column(DECIMAL(10, 5))
    MWFLOW = Column(DECIMAL(10, 5))
    MWLOSSES = Column(DECIMAL(10, 5))
    EXPORTLIMIT = Column(DECIMAL(10, 5))
    IMPORTLIMIT = Column(DECIMAL(10, 5))
    MARGINALLOSS = Column(DECIMAL(6, 5))
    FCASEXPORTLIMIT = Column(DECIMAL(10, 5))
    FCASIMPORTLIMIT = Column(DECIMAL(10, 5))

    INTERCONNECTORID1 = relationship("INTERCONNECTORID")


class PREDISPATCHINTERCONNECTORSOLN(Base):
    __tablename__ = "PREDISPATCH_INTERCONNECTOR_SOLN"
    __table_args__ = (
        Index(
            "uniq",
            "INTERCONNECTORID",
            "DATETIME",
            "PREDISPATCHSEQNO",
            "INTERVENTION",
            unique=True,
        ),
    )

    id = Column(INTEGER(11), primary_key=True)
    PREDISPATCHSEQNO = Column(INTEGER(11), index=True)
    INTERCONNECTORID = Column(
        ForeignKey("nemweb_meta.INTERCONNECTORID.id"), index=True
    )
    PERIODID = Column(TINYINT(4), index=True)
    METEREDMWFLOW = Column(DECIMAL(10, 5))
    MWFLOW = Column(DECIMAL(10, 5))
    MWLOSSES = Column(DECIMAL(10, 5))
    DATETIME = Column(DateTime, index=True)
    EXPORTLIMIT = Column(DECIMAL(10, 5))
    IMPORTLIMIT = Column(DECIMAL(10, 5))
    MARGINALLOSS = Column(DECIMAL(6, 5), nullable=False)
    INTERVENTION = Column(TINYINT(4), nullable=False, index=True)

    INTERCONNECTORID1 = relationship("INTERCONNECTORID")


class TRADINGINTERCONNECTORRE(Base):
    __tablename__ = "TRADING_INTERCONNECTORRES"
    __table_args__ = (
        Index("uniq", "SETTLEMENTDATE", "INTERCONNECTORID", unique=True),
    )

    id = Column(INTEGER(11), primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    INTERCONNECTORID = Column(
        ForeignKey("nemweb_meta.INTERCONNECTORID.id"), index=True
    )
    METEREDMWFLOW = Column(DECIMAL(6, 2))
    MWFLOW = Column(DECIMAL(6, 2))
    MWLOSSES = Column(DECIMAL(6, 2))

    INTERCONNECTORID1 = relationship("INTERCONNECTORID")



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, validator

from opennem.api.photo.schema import Photo
from opennem.api.stats.schema import OpennemData
from opennem.api.weather.schema import WeatherStation
from opennem.core.dispatch_type import DispatchType
from opennem.core.normalizers import (  # clean_numbers,; station_name_cleaner,
    clean_capacity,
    normalize_string,
)

from .core import BaseConfig
from .network import NetworkNEM, NetworkSchema


class OpennemBaseSchema(BaseConfig):

    created_by: Optional[str]
    created_at: Optional[datetime] = datetime.now()


class FueltechSchema(BaseConfig):
    code: str
    label: str
    renewable: bool


class FacilityStatusSchema(BaseConfig):
    code: str
    label: str


class ParticipantSchema(OpennemBaseSchema):
    id: int
    code: Optional[str]
    name: str
    network_name: Optional[str] = None
    network_code: Optional[str] = None
    country: Optional[str] = None
    abn: Optional[str] = None


class StationBaseSchema(OpennemBaseSchema):
    id: int


class FacilityBaseSchema(OpennemBaseSchema):
    id: int


class GeoPoint(BaseModel):
    lat: float
    lng: float


class GeoBoundary(BaseModel):
    pass


class RecordTypes(str, Enum):
    station = "station"
    facility = "facility"
    location = "location"
    revision = "revision"


class RevisionSchema(OpennemBaseSchema):
    id: int

    changes: Dict[str, Union[str, int, float, bool, None]] = {}
    previous: Optional[Dict[str, Union[str, int, float, bool]]] = {}

    parent_id: Optional[int]
    parent_type: Optional[str]
    station_owner_id: Optional[int]
    station_owner_name: Optional[str]
    station_owner_code: Optional[str]

    is_update: bool = False

    approved: bool = False
    approved_by: Optional[str]
    approved_at: Optional[datetime]
    approved_comment: Optional[str]

    discarded: bool = False
    discarded_by: Optional[str]
    discarded_at: Optional[datetime]


class ScadaReading(Tuple[datetime, Optional[float]]):
    pass


class FacilitySchema(OpennemBaseSchema):
    id: Optional[int]

    network: NetworkSchema = NetworkNEM

    fueltech: Optional[FueltechSchema]

    status: Optional[FacilityStatusSchema]

    station_id: Optional[int]

    # @TODO no longer optional
    code: Optional[str] = ""

    scada_power: Optional[OpennemData]

    # revisions: Optional[List[RevisionSchema]] = []
    # revision_ids: Optional[List[int]] = []

    dispatch_type: DispatchType = "GENERATOR"

    active: bool = True

    capacity_registered: Optional[float]

    registered: Optional[datetime]
    deregistered: Optional[datetime]

    network_region: Optional[str]

    unit_id: Optional[int]
    unit_number: Optional[int]
    unit_alias: Optional[str]
    unit_capacity: Optional[float]

    approved: bool = False
    approved_by: Optional[str]
    approved_at: Optional[datetime]

    @validator("capacity_registered")
    def _clean_capacity_regisered(cls, value):
        value = clean_capacity(value)

        if isinstance(value, float):
            value = round(value, 2)

        return value


class WeatherStationNearest(BaseModel):
    code: str
    distance: float


class LocationSchema(OpennemBaseSchema):
    id: Optional[int]

    weather_station: Optional[WeatherStation]
    weather_nearest: Optional[WeatherStationNearest]

    address1: Optional[str] = ""
    address2: Optional[str] = ""
    locality: Optional[str] = ""
    state: Optional[str] = ""
    postcode: Optional[str] = ""
    country: Optional[str] = "au"

    # Geo fields
    # place_id: Optional[str]
    geocode_approved: bool = False
    geocode_skip: bool = False
    geocode_processed_at: Optional[datetime] = None
    geocode_by: Optional[str]
    # geom: Optional[WKBElement] = None
    # boundary: Optional[WKBElement] = None

    # geo fields
    lat: Optional[float]
    lng: Optional[float]

    @validator("address1")
    def clean_address(cls, value):
        return normalize_string(value)

    @validator("address2")
    def clean_address2(cls, value):
        return normalize_string(value)

    @validator("locality")
    def clean_locality(cls, value):
        return normalize_string(value)

    @validator("state")
    def state_upper(cls, value: str) -> Optional[str]:
        if value:
            return value.upper()

    @validator("postcode")
    def clean_postcode(cls, value: str) -> Optional[str]:
        if value:
            return value.strip()

    @property
    def geom(self) -> Optional[str]:
        if not self.lng or not self.lat:
            return None

        geom = "SRID=4326;POINT({} {})".format(self.lng, self.lat,)

        return geom


class StationSchema(OpennemBaseSchema):
    id: Optional[int]

    participant: Optional[ParticipantSchema] = None
    participant_id: Optional[str]

    facilities: Optional[List[FacilitySchema]] = []

    photos: Optional[List[Photo]]

    # history: Optional[List[__self__]]

    # revisions: Optional[List[RevisionSchema]]

    code: str

    name: Optional[str]

    # Original network fields
    network_name: Optional[str]

    location: LocationSchema = LocationSchema()
    location_id: Optional[int]

    approved: bool = False
    approved_by: Optional[str]
    approved_at: Optional[datetime]

    network: Optional[NetworkSchema] = None

    description: Optional[str]
    wikipedia_link: Optional[str]
    wikidata_id: Optional[str]




import csv
import logging

from opennem.core.normalizers import (
    normalize_duid,
    normalize_string,
    participant_name_filter,
)
from opennem.db.models.opennem import Participant
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)

# @TODO we can probably merge these two pipelines now


class WemStoreParticipant(DatabaseStoreBase):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        record_count = 0
        csvreader = csv.DictReader(item["content"].split("\n"))

        for row in csvreader:
            if "Participant Code" not in row:
                logger.error("Invalid row")
                continue

            participant = None

            participant_code = normalize_duid(row["Participant Code"])
            participant_name = participant_name_filter(row["Participant Name"])

            participant = (
                s.query(Participant)
                .filter(Participant.code == participant_code)
                .one_or_none()
            )

            if not participant:
                participant = Participant(
                    code=participant_code,
                    name=participant_name,
                    # @TODO WEM provides these but nem doesn't so ignore for noe
                    # address=row["Address"],
                    # city=row["City"],
                    # state=row["State"],
                    # postcode=row["Postcode"],
                    created_by="pipeline.wem.participant",
                )

                logger.info(
                    "Created new WEM participant: {}".format(participant_code)
                )

            elif participant.name != participant_name:
                participant.name = participant_name
                participant.updated_by = "pipeline.wem.participant"
                logger.info(
                    "Updated WEM participant: {}".format(participant_code)
                )

            try:
                s.add(participant)
                s.commit()
                record_count += 1
            except Exception as e:
                logger.error(e)

        s.close()

        return record_count


class WemStoreLiveParticipant(DatabaseStoreBase):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        record_count = 0
        csvreader = csv.DictReader(item["content"].split("\n"))

        for row in csvreader:
            if "PARTICIPANT_CODE" not in row:
                logger.error("Invalid row")
                continue

            participant = None

            participant_code = normalize_duid(
                normalize_string(row["PARTICIPANT_CODE"])
            )
            participant_name = participant_name_filter(
                row["PARTICIPANT_DISPLAY_NAME"]
            ) or participant_name_filter(
                row.get("PARTICIPANT_FULL_NAME", None)
            )

            participant = (
                s.query(Participant)
                .filter(Participant.code == participant_code)
                .one_or_none()
            )

            if not participant:
                participant = Participant(
                    code=participant_code,
                    name=participant_name,
                    created_by=spider.name,
                )

                logger.info(
                    "Created new WEM participant: {}".format(participant_code)
                )

            elif participant.name != participant_name:
                participant.name = participant_name
                participant.updated_by = spider.name

                logger.info(
                    "Updated WEM participant: {}".format(participant_code)
                )

            try:
                s.add(participant)
                s.commit()
                record_count += 1
            except Exception as e:
                logger.error(e)

        s.close()

        return record_count



import csv
import logging
from datetime import datetime, timedelta

from sqlalchemy.dialects.postgresql import insert

from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionAutocommit, SessionLocal, get_database_engine
from opennem.db.models.opennem import FacilityScada
from opennem.pipelines import DatabaseStoreBase
from opennem.schema.network import NetworkWEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def generate_records(csvreader, spider):
    for row in csvreader:
        trading_interval = parse_date(
            row["Trading Interval"], network=NetworkWEM, dayfirst=False
        )
        facility_code = normalize_duid(row["Facility Code"])

        generated = 0.0
        generated_field = "EOI Quantity (MW)"

        if generated_field in row:
            generated = clean_float(row[generated_field])

        energy = 0.0
        energy_field = "Energy Generated (MWh)"

        if energy_field in row:
            energy = clean_float(row[energy_field])

        yield {
            "created_by": spider.name,
            # "updated_by": None,
            "network_id": "WEM",
            "trading_interval": trading_interval,
            "facility_code": facility_code,
            "generated": generated,
            "eoi_quantity": energy,
        }


class WemStoreFacilityScada(DatabaseStoreBase):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        session = SessionAutocommit()
        engine = get_database_engine()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_stored = 0

        for record in generate_records(csvreader, spider):
            stmt = insert(FacilityScada).values(record)
            stmt.bind = engine
            stmt = stmt.on_conflict_do_update(
                constraint="facility_scada_pkey",
                set_={
                    # "updated_by": stmt.excluded.created_by,
                    "eoi_quantity": stmt.excluded.eoi_quantity,
                    "generated": stmt.excluded.generated,
                },
            )

            try:
                session.execute(stmt)
                records_stored += 1
            except Exception as e:
                logger.error("Error: {}".format(e))
            finally:
                session.close()

        return records_stored


class WemStoreLiveFacilityScada(DatabaseStoreBase):
    """
        Store live facility scada data.

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        session = SessionLocal()
        engine = get_database_engine()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_to_store = []
        last_asat = None

        for row in csvreader:

            # @TODO MAX_GEN_CAPACITY
            # facility_capacity = row["MAX_GEN_CAPACITY"]

            if row["AS_AT"] != "":
                last_asat = parse_date(
                    row["AS_AT"], network=NetworkWEM, dayfirst=False
                )

            if not last_asat or type(last_asat) is not datetime:
                logger.error("Invalid row or no datetime")
                continue

            # We need to pivot the table since columns are time intervals
            for i in range(1, 48):

                column = f"I{i:02}"

                if column not in row:
                    logger.error(
                        "Do not have data for interval {}".format(column)
                    )
                    continue

                if i > 0:
                    interval = last_asat - timedelta(minutes=(i - 1) * 30)
                else:
                    interval = last_asat

                facility_code = normalize_duid(row["FACILITY_CODE"])

                val = None

                try:
                    val = float(row[column]) / 2 or None
                except ValueError:
                    pass

                records_to_store.append(
                    {
                        "created_by": spider.name,
                        "network_id": "WEM",
                        "trading_interval": interval,
                        "facility_code": facility_code,
                        "eoi_quantity": val,
                    }
                )

        stmt = insert(FacilityScada).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            constraint="facility_scada_pkey",
            set_={
                # "updated_by": stmt.excluded.created_by,
                "eoi_quantity": stmt.excluded.eoi_quantity,
            },
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
        finally:
            session.close()

        return len(records_to_store)


class WemStoreFacilityIntervals(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        session = SessionLocal()
        engine = get_database_engine()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_to_store = []

        for row in csvreader:
            trading_interval = parse_date(
                row["PERIOD"], network=NetworkWEM, dayfirst=False
            )
            facility_code = normalize_duid(row["FACILITY_CODE"])

            energy = 0.0

            if "POTENTIAL_MWH" in row:
                energy = clean_float(row["POTENTIAL_MWH"])

                # @NOTE - don't use
                energy = 0.0

            records_to_store.append(
                {
                    "created_by": spider.name,
                    "network_id": "WEM",
                    "trading_interval": trading_interval,
                    "facility_code": facility_code,
                    "generated": row["ACTUAL_MW"] or None,
                    "eoi_quantity": energy,
                }
            )

        stmt = insert(FacilityScada).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            constraint="facility_scada_pkey",
            set_={
                # "updated_by": stmt.excluded.created_by,
                "generated": stmt.excluded.generated,
                "eoi_quantity": stmt.excluded.eoi_quantity,
            },
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
        finally:
            session.close()

        return len(records_to_store)



from opennem.schema.network import NetworkNEM, NetworkSchema, NetworkWEM


def network_from_state(state: str) -> NetworkSchema:
    state = state.upper().strip()

    if state in ["WA"]:
        return NetworkWEM

    if state in ["QLD", "NSW", "VIC", "ACT", "TAS", "SA", "NT"]:
        return NetworkNEM

    raise Exception("Unknown network {}".format(state))


def network_from_network_region(network_region: str,) -> NetworkSchema:
    network_region = network_region.upper()

    if network_region in ["WEM", "WA1"]:
        return NetworkWEM
    if network_region in ["NEM", "NSW1", "QLD1", "SA1", "VIC1", "TAS1"]:
        return NetworkNEM

    raise Exception("Unknown network {}".format(network_region))


def network_from_network_code(network_code: str) -> NetworkSchema:
    network_code = network_code.upper().strip()

    if network_code in ["WEM"]:
        return NetworkWEM

    if network_code in ["NEM"]:
        return NetworkNEM

    raise Exception("Unknown network {}".format(network_code))



from huey import RedisHuey, crontab

from opennem.api.exporter import wem_run_all
from opennem.exporter.geojson import export_facility_geojson
from opennem.settings import settings

huey = RedisHuey("opennem.exporter", host=settings.cache_url.host)


@huey.periodic_task(crontab(minute="*/5"))
def wem_export_task():
    wem_run_all()


@huey.periodic_task(crontab(minute="*/5"))
def export_geojson():
    export_facility_geojson()



from fastapi import APIRouter, Depends, HTTPException
from fastapi.param_functions import Query
from geojson_pydantic.features import FeatureCollection
from sqlalchemy.orm.session import Session
from starlette import status

from opennem.api.geo.schema import FacilityGeo
from opennem.api.station.controllers import get_stations
from opennem.db import get_database_session

from .controllers import stations_to_geojson

router = APIRouter()


@router.get(
    "/facilities", name="Facility Geo", response_model=FacilityGeo,
)
def geo_facilities_api(
    only_approved: bool = Query(
        True, description="Only show approved stations"
    ),
    session: Session = Depends(get_database_session),
) -> FacilityGeo:
    stations = get_stations(session, only_approved=only_approved)

    if not stations:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No stations found"
        )

    stations_geo = stations_to_geojson(stations)

    return stations_geo



from sqlalchemy.orm import sessionmaker

from opennem.db import db_connect

engine = db_connect()
session = sessionmaker(bind=engine)


PERIOD_MAP = {
    "all": 0,
    # current year
    "year": "date_trunc('year', current_date))",
    # current month
    "month": "date_trunc('month', current_date))",
}


def bom_observation(station_id, region="wa"):
    __query = """
        SELECT
            observation_time,
            round(temp_air, 2) as value
        FROM bom_observation
        WHERE
            station_id = '{station_id}'
            AND observation_time > now() - interval '7 days'
        order by 1 desc
    """

    query = __query.format(station_id=station_id)

    json_object = {
        "id": f"{region}.temperature",
        "history": {
            "data": [],
            "interval": "30m",
            "last": None,
            "start": None,
        },
        "bom.station": station_id,
        "region": region,
        "type": "temperature",
        "units": "degreesC",
    }

    with engine.connect() as c:

        rows = c.execute(query)

        for row in rows:

            if (
                json_object["history"]["start"] == None
                or row[0] < json_object["history"]["start"]
            ):
                json_object["history"]["start"] = row[0]

            if (
                json_object["history"]["last"] == None
                or row[0] > json_object["history"]["last"]
            ):
                json_object["history"]["last"] = row[0]

            json_object["history"]["data"].append(row[1])

    return json_object


def wem_price(region="wa"):
    __query = """
        with intervals as (
        select generate_series(
            date_trunc('hour', now()) - '7 day'::interval,
            date_trunc('hour', now()),
            '30 minutes'::interval
        ) as interval
        )

        select
            i.interval,
            round(bs.price, 2),
            case when bs.price is null then 0 else bs.price end as price_cast,
            round(bs.generation_total, 2)
        from intervals i
        left join balancing_summary bs on bs.trading_interval = interval
        where bs.network_id='WEM'
        order by i.interval desc
    """

    query = __query.format()

    json_object = {
        "id": f"{region}.price",
        "history": {
            "data": [],
            "interval": "30m",
            "last": None,
            "start": None,
        },
        "region": region,
        "type": "price",
        "units": "AUD/MWh",
    }

    with engine.connect() as c:

        rows = c.execute(query)

        for row in rows:

            if (
                json_object["history"]["start"] == None
                or row[0] < json_object["history"]["start"]
            ):
                json_object["history"]["start"] = row[0]

            if (
                json_object["history"]["last"] == None
                or row[0] > json_object["history"]["last"]
            ):
                json_object["history"]["last"] = row[0]

            json_object["history"]["data"].append(row[1])

    return json_object


def wem_demand(region="wa"):
    __query = """
        with intervals as (
            select generate_series(
                date_trunc('hour', now()) - '7 day'::interval,
                date_trunc('hour', now()),
                '30 minutes'::interval
            ) as interval
        )

        select
            i.interval,
            round(wbs.price, 2),
            round(wbs.generation_total, 2)
        from intervals i
        left join balancing_summary wbs on wbs.trading_interval = interval
        where wbs.network_id='WEM'
        order by i.interval desc
    """

    query = __query.format()

    json_object = {
        "id": f"{region}.demand",
        "history": {
            "data": [],
            "interval": "30m",
            "last": None,
            "start": None,
        },
        "region": region,
        "type": "demand",
        "units": "MW",
    }

    with engine.connect() as c:

        rows = c.execute(query)

        for row in rows:

            if (
                json_object["history"]["start"] == None
                or row[0] < json_object["history"]["start"]
            ):
                json_object["history"]["start"] = row[0]

            if (
                json_object["history"]["last"] == None
                or row[0] > json_object["history"]["last"]
            ):
                json_object["history"]["last"] = row[0]

            json_object["history"]["data"].append(row[2])

    return json_object


def wem_power_groups(intervals_per_hour=2):
    __query = """
        with intervals as (
            select generate_series(
                date_trunc('hour', now()) - '7 day'::interval,
                date_trunc('hour', now()),
                '30 minutes'::interval
            ) as interval
        )
        select
            i.interval,
            f.fueltech_id as fueltech,
            round(sum(fs.generated) * {intervals_per_hour}, 2) as val
        from intervals i
        left join facility_scada fs on i.interval = fs.trading_interval
        left join facility f on f.code = fs.facility_code
        where
            fs.network_id='WEM'
            and f.fueltech_id is not null
        group by 1, 2
        order by 2 asc, 1 desc
        """

    query = __query.format(intervals_per_hour=intervals_per_hour)

    json_envelope = []
    json_obj = None

    with engine.connect() as c:
        rows = c.execute(query)

        current_tech = None

        for row in rows:

            if current_tech != row[1]:

                if json_obj is not None:
                    json_envelope.append(json_obj)

                current_tech = row[1]

                json_obj = {
                    "id": f"wem.fuel_tech.{current_tech}.power",
                    "fuel_tech": current_tech,
                    "region": "wa",
                    "network": "wem",
                    "type": "power",
                    "units": "MWh",
                    "history": {
                        "interval": "30m",
                        "start": None,
                        "last": None,
                        "data": [],
                    },
                }

            if (
                json_obj["history"]["start"] == None
                or row[0] < json_obj["history"]["start"]
            ):
                json_obj["history"]["start"] = row[0]

            if (
                json_obj["history"]["last"] == None
                or row[0] > json_obj["history"]["last"]
            ):
                json_obj["history"]["last"] = row[0]

            value = row[2]

            json_obj["history"]["data"].append(value)

    json_envelope.append(json_obj)

    return json_envelope


def wem_energy_year(year="2020"):
    """
        WEM energy year

    """
    __query = """
        select
            date_trunc('day', fs.trading_interval) AS trading_day,
            round(max(fs.eoi_quantity), 2) as energy_output,
            f.fueltech_id
        from facility_scada fs
        left join facility f on fs.facility_code = f.code
        where
            f.fueltech_id is not null
            and f.network_id = 'WEM'
            and extract('year' from fs.trading_interval) = {year}
        group by 1, f.fueltech_id
        order by 1 desc, 2 asc
        """

    query = __query.format(year=year)

    json_envelope = {}

    with engine.connect() as c:
        rows = c.execute(query)

        current_tech = None

        for row in rows:

            current_tech = row[2]

            if current_tech not in json_envelope.keys():
                json_envelope[current_tech] = {
                    "id": f"wem.fuel_tech.{current_tech}.energy",
                    "fuel_tech": current_tech,
                    "region": "wa",
                    "type": "energy",
                    "units": "MWh",
                    "history": {
                        "interval": "1d",
                        "start": None,
                        "last": None,
                        "data": [],
                    },
                }

            if (
                json_envelope[current_tech]["history"]["start"] == None
                or row[0] < json_envelope[current_tech]["history"]["start"]
            ):
                json_envelope[current_tech]["history"]["start"] = row[0]

            if (
                json_envelope[current_tech]["history"]["last"] == None
                or row[0] > json_envelope[current_tech]["history"]["last"]
            ):
                json_envelope[current_tech]["history"]["last"] = row[0]

            json_envelope[current_tech]["history"]["data"].append(row[1])

    return [json_envelope[i] for i in json_envelope.keys()]


def wem_market_value_year(year="2020"):
    """
        WEM market value year

    """
    __query = """
        select
            date_trunc('day', fs.trading_interval) AS trading_day,
            round(sum(fs.eoi_quantity * bs.price), 2) as energy_interval,
            f.fueltech_id
        from facility_scada fs
        left join facility f on fs.facility_code = f.code
        join balancing_summary bs on fs.trading_interval = bs.trading_interval
        where
            f.fueltech_id is not null
            and fs.network_id='WEM'
            and extract('year' from fs.trading_interval) = {year}
        group by 1, f.fueltech_id
        order by 1 desc, 2 asc
        """

    query = __query.format(year=year)

    json_envelope = {}

    with engine.connect() as c:
        rows = c.execute(query)

        current_tech = None

        for row in rows:

            current_tech = row[2]

            if current_tech not in json_envelope.keys():
                json_envelope[current_tech] = {
                    "id": f"wem.fuel_tech.{current_tech}.market_value",
                    "fuel_tech": current_tech,
                    "region": "wa",
                    "type": "market_value",
                    "units": "AUD",
                    "history": {
                        "interval": "1d",
                        "start": None,
                        "last": None,
                        "data": [],
                    },
                }

            if (
                json_envelope[current_tech]["history"]["start"] == None
                or row[0] < json_envelope[current_tech]["history"]["start"]
            ):
                json_envelope[current_tech]["history"]["start"] = row[0]

            if (
                json_envelope[current_tech]["history"]["last"] == None
                or row[0] > json_envelope[current_tech]["history"]["last"]
            ):
                json_envelope[current_tech]["history"]["last"] = row[0]

            json_envelope[current_tech]["history"]["data"].append(row[1])

    return [json_envelope[i] for i in json_envelope.keys()]


def wem_energy_all():
    """
        WEM energy year

    """
    __query = """
        select
            date_trunc('month', fs.trading_interval) AS trading_day,
            round(max(fs.eoi_quantity), 2) as energy_output,
            f.fueltech_id
        from facility_scada fs
        left join facility f on fs.facility_code = f.code
        where
            f.fueltech_id is not null
            and f.network_id='WEM'
        group by 1, f.fueltech_id
        order by 1 desc, 2 asc
        """

    query = __query.format()

    json_envelope = {}

    with engine.connect() as c:
        rows = c.execute(query)

        current_tech = None

        for row in rows:

            current_tech = row[2]

            if current_tech not in json_envelope.keys():
                json_envelope[current_tech] = {
                    "id": f"wem.fuel_tech.{current_tech}.energy",
                    "fuel_tech": current_tech,
                    "region": "wa",
                    "type": "energy",
                    "units": "MWh",
                    "history": {
                        "interval": "1M",
                        "start": None,
                        "last": None,
                        "data": [],
                    },
                }

            if (
                json_envelope[current_tech]["history"]["start"] == None
                or row[0] < json_envelope[current_tech]["history"]["start"]
            ):
                json_envelope[current_tech]["history"]["start"] = row[0]

            if (
                json_envelope[current_tech]["history"]["last"] == None
                or row[0] > json_envelope[current_tech]["history"]["last"]
            ):
                json_envelope[current_tech]["history"]["last"] = row[0]

            json_envelope[current_tech]["history"]["data"].append(row[1])

    return [json_envelope[i] for i in json_envelope.keys()]


def wem_market_value_all():
    """
        WEM market value all

    """
    __query = """
        select
            date_trunc('month', fs.trading_interval) AS trading_day,
            round(sum(fs.eoi_quantity * bs.price), 2) as energy_interval,
            f.fueltech_id
        from facility_scada fs
        left join facility f on fs.facility_code = f.code
        join balancing_summary bs on fs.trading_interval = bs.trading_interval
        where
            f.fueltech_id is not null
            and f.network_id='WEM'
        group by 1, f.fueltech_id
        order by 1 desc, 2 asc
        """

    query = __query.format()

    json_envelope = {}

    with engine.connect() as c:
        rows = c.execute(query)

        current_tech = None

        for row in rows:

            current_tech = row[2]

            if current_tech not in json_envelope.keys():
                json_envelope[current_tech] = {
                    "id": f"wem.fuel_tech.{current_tech}.market_value",
                    "fuel_tech": current_tech,
                    "region": "wa",
                    "type": "market_value",
                    "units": "AUD",
                    "history": {
                        "interval": "1M",
                        "start": None,
                        "last": None,
                        "data": [],
                    },
                }

            if (
                json_envelope[current_tech]["history"]["start"] == None
                or row[0] < json_envelope[current_tech]["history"]["start"]
            ):
                json_envelope[current_tech]["history"]["start"] = row[0]

            if (
                json_envelope[current_tech]["history"]["last"] == None
                or row[0] > json_envelope[current_tech]["history"]["last"]
            ):
                json_envelope[current_tech]["history"]["last"] = row[0]

            json_envelope[current_tech]["history"]["data"].append(row[1])

    return [json_envelope[i] for i in json_envelope.keys()]
